local cartridge = require('cartridge')
local checks = require('checks')
local vshard = require('vshard')

---@class (exact) Customer
---@field customer_id number
---@field name string

---@class (exact) Account
---@field customer_id number
---@field account_id number
---@field balance number
---@field name string

---@param customer Customer
---@return any
local function add_customer(customer)
    checks({customer_id = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(customer.customer_id)
    print(bucket_id)
    return vshard.router.callrw(bucket_id, 'add_customer', { bucket_id, customer })
end

---@param account Account
---@return any
local function add_account(account)
    checks({account_id = 'number', customer_id = 'number', balance = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(account.customer_id)
    return vshard.router.callrw(bucket_id, 'add_account', { bucket_id, account })
end

---@param account Account
---@return any
local function add_account_sync(account)
    checks({account_id = 'number', customer_id = 'number', balance = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(account.customer_id)
    return vshard.router.callrw(bucket_id, 'add_account_sync', { bucket_id, account })
end

---@param customer Customer
---@return any
local function update_customer(customer)
    checks({customer_id = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(customer.customer_id)
    return vshard.router.callrw(bucket_id, 'update_customer', { bucket_id, customer })
end

---@param account Account
---@return any
local function update_account(account)
    checks({account_id = 'number', customer_id = 'number', balance = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(account.customer_id)
    return vshard.router.callrw(bucket_id, 'update_account', { bucket_id, account })
end

---@param account Account
---@return any
local function update_account_sync(account)
    checks({account_id = 'number', customer_id = 'number', balance = 'number', name = 'string'})
    local bucket_id = vshard.router.bucket_id(account.customer_id)
    return vshard.router.callrw(bucket_id, 'update_account_sync', { bucket_id, account })
end

---@param customer_id number
---@return any
local function delete_customer(customer_id)
    checks('number')
    local bucket_id = vshard.router.bucket_id(customer_id)
    return vshard.router.callrw(bucket_id, 'delete_customer', { customer_id })
end

---@param customer_id number
---@param account_id number
---@return any
local function delete_account(customer_id, account_id)
    checks('number', 'number')
    local bucket_id = vshard.router.bucket_id(customer_id)
    return vshard.router.callrw(bucket_id, 'delete_account', { account_id })
end

---@param customer_id number
---@param account_id number
---@return any
local function delete_account_sync(customer_id, account_id)
    checks('number', 'number')
    local bucket_id = vshard.router.bucket_id(customer_id)
    return vshard.router.callrw(bucket_id, 'delete_account_sync', { account_id })
end

local function init(opts) -- luacheck: no unused args
    box.once("create_user", function()
        box.schema.user.grant('guest', 'super')
        box.schema.user.create('uroboros', { password = 'password' })
        box.schema.user.grant('uroboros', 'read,write,create,alter,drop,execute,session', 'universe')
    end)

    rawset(_G, 'add_customer', add_customer)
    rawset(_G, 'add_account', add_account)
    rawset(_G, 'add_account_sync', add_account_sync)

    rawset(_G, 'update_customer', update_customer)
    rawset(_G, 'update_account', update_account)
    rawset(_G, 'update_account_sync', update_account_sync)

    rawset(_G, 'delete_customer', delete_customer)
    rawset(_G, 'delete_account', delete_account)
    rawset(_G, 'delete_account_sync', delete_account_sync)


    local httpd = assert(cartridge.service_get('httpd'), "Failed to get httpd service")
    local json = require('json')
    local add_customer_http = function (req)
        local customer = req:json()

        local res = add_customer(customer)
        return { body = json.encode(res), status = 200 }
    end
    httpd:route({method = 'POST', path = '/customer'}, add_customer_http)
    return true
end

local function stop()
    return true
end

local function validate_config(conf_new, conf_old) -- luacheck: no unused args
    return true
end

local function apply_config(conf, opts) -- luacheck: no unused args
    return true
end

return {
    role_name = 'app.roles.router',
    init = init,
    stop = stop,
    validate_config = validate_config,
    apply_config = apply_config,
    dependencies = {'uroboros.roles.topology'},
    utils = {
        add_customer = add_customer,
        add_account = add_account,
        add_account_sync = add_account_sync,
        update_customer = update_customer,
        update_account = update_account,
        update_account_sync = update_account_sync,
        delete_customer = delete_customer,
        delete_account = delete_account,
        delete_account_sync = delete_account_sync
    }
}
