local checks = require('checks')

local function add_customer(bucket_id, customer)
    checks('number', {customer_id = 'number', name = 'string'})
    box.space.customer:insert {
        customer.customer_id,
        bucket_id,
        customer.name
    }
end

local function add_account(bucket_id, account)
    checks('number', {
        account_id = 'number',
        customer_id = 'number',
        balance = 'number',
        name = 'string'
    })
    box.space.account:insert {
        account.account_id,
        account.customer_id,
        bucket_id,
        account.balance,
        account.name
    }
end

local function add_account_sync(bucket_id, account)
    checks('number', {
        account_id = 'number',
        customer_id = 'number',
        balance = 'number',
        name = 'string'
    })
    box.space.account_sync:insert {
        account.account_id,
        account.customer_id,
        bucket_id,
        account.balance,
        account.name
    }
end

local function update_customer(bucket_id, customer)
    checks('number', {customer_id = 'number', name = 'string'})
    box.space.customer:replace {
        customer.customer_id,
        bucket_id,
        customer.name
    }
end

local function update_account(bucket_id, account)
    checks('number', {
        account_id = 'number',
        customer_id = 'number',
        balance = 'number',
        name = 'string'
    })
    box.space.account:replace {
        account.account_id,
        account.customer_id,
        bucket_id,
        account.balance,
        account.name
    }
end

local function update_account_sync(bucket_id, account)
    checks('number', {
        account_id = 'number',
        customer_id = 'number',
        balance = 'number',
        name = 'string'
    })
    box.space.account_sync:replace {
        account.account_id,
        account.customer_id,
        bucket_id,
        account.balance,
        account.name
    }
end

local function delete_customer(customer_id)
    checks('number')
    box.space.customer:delete { customer_id }
end

local function delete_account(account_id)
    checks('number')
    box.space.account:delete { account_id }
end

local function delete_account_sync(account_id)
    checks('number')
    box.space.account_sync:delete { account_id }
end

local function init(opts) -- luacheck: no unused args
    box.once("create_user", function()
        box.schema.user.grant('guest', 'super')
        box.schema.user.create('uroboros', { password = 'password' })
        box.schema.user.grant('uroboros', 'read,write,create,alter,drop,execute,session', 'universe')
    end)

    if opts.is_master then
        local customer = box.schema.space.create('customer', {
            if_not_exists = true,
            format = {
                { 'customer_id', 'unsigned' },
                { 'bucket_id',   'unsigned' },
                { 'name',        'string' },
            }
        })
        customer:create_index('customer_id', { parts = { {'customer_id'} }, if_not_exists = true })
        customer:create_index('bucket_id', { parts = {
            { 'bucket_id', is_nullable = false}
        }, unique = false, if_not_exists = true })

        local account = box.schema.space.create('account', {
            if_not_exists = true,
            format = {
                { 'account_id',  'unsigned' },
                { 'customer_id', 'unsigned' },
                { 'bucket_id',   'unsigned' },
                { 'balance',     'unsigned' },
                { 'name',        'string' },
            }
        })
        account:create_index('account_id', { parts = { {'account_id'} }, if_not_exists = true })
        account:create_index('customer_id', { parts = { {'customer_id'} }, unique = false, if_not_exists = true })
        account:create_index('bucket_id', { parts = { {'bucket_id'} }, unique = false, if_not_exists = true })

        local account_sync = box.schema.space.create('account_sync', {
            is_sync = true,
            if_not_exists = true,
            format = {
                { 'account_id',  'unsigned' },
                { 'customer_id', 'unsigned' },
                { 'bucket_id',   'unsigned' },
                { 'balance',     'unsigned' },
                { 'name',        'string' },
            }
        })
        account_sync:create_index('account_id', { parts = { {'account_id'} }, if_not_exists = true })
        account_sync:create_index('customer_id', { parts = { {'customer_id'} }, unique = false, if_not_exists = true })
        account_sync:create_index('bucket_id', { parts = { {3, 'unsigned'} }, unique = false, if_not_exists = true })

        box.ctl.promote()
    end

    rawset(_G, 'add_customer', add_customer)
    rawset(_G, 'add_account', add_account)
    rawset(_G, 'add_account_sync', add_account_sync)

    rawset(_G, 'update_customer', update_customer)
    rawset(_G, 'update_account', update_account)
    rawset(_G, 'update_account_sync', update_account_sync)

    rawset(_G, 'delete_customer', delete_customer)
    rawset(_G, 'delete_account', delete_account)
    rawset(_G, 'delete_account_sync', delete_account_sync)

    return true
end

local function stop()
    return true
end

local function validate_config(conf_new, conf_old) -- luacheck: no unused args
    return true
end

local function apply_config(conf, opts) -- luacheck: no unused args
    return true
end

return {
    role_name = 'app.roles.storage_left',
    init = init,
    stop = stop,
    validate_config = validate_config,
    apply_config = apply_config,
    dependencies = {'uroboros.roles.replicator'},
    utils = {
        add_customer = add_customer,
        add_account = add_account,
        add_account_sync = add_account_sync,
        update_customer = update_customer,
        update_account = update_account,
        update_account_sync = update_account_sync,
        delete_customer = delete_customer,
        delete_account = delete_account,
        delete_account_sync = delete_account_sync
    }
}
