from collections import defaultdict
from contextlib import suppress
import io
import os
import requests
import shutil
import signal
import subprocess
import sys
import time
import threading

import pytest
from typing import (
    Any,
    Generator,
    Optional,
    Tuple,
)
import tarantool
from tarantool.error import (  # type: ignore
    tnt_strerror,
    DatabaseError,
)


def format_format(format: list[dict[str, str]]) -> str:
    items = []
    for item in format:
            name = item["name"]
            kind = item["type"]
            s = f"name='{name}', type='{kind}'"
            items.append("{" + s + "}")
    return "{" + ", ".join(items) + "}"


_1_SEC = 1
BASE_HOST = "localhost"
BASE_USER = "uroboros"
BASE_PASSWORD = "password"
PICO_BINARY_PATH = os.environ.get("PICODATA_BINARY_PATH") or "../picodata/target/debug/picodata"
PICO_SERVICE_USER = "pico_service"
TEST_RUN_CLUSTERS = os.environ.get("TEST_RUN_CLUSTERS") or "docker"


OUT_LOCK = threading.Lock()



class ErrorCode:
    Loading = 116
    NoSuchInstance = 10016
    NoSuchReplicaset = 10017


def eprint(*args: Any, **kwargs: Any):
    print(*args, file=sys.stderr, **kwargs)


class TarantoolError(Exception):
    """
    Raised when Tarantool responds with an IPROTO_ERROR.
    """

    ...


class ReturnError(Exception):
    """
    Raised when Tarantool returns `nil, err`.
    """

    ...


class MalformedAPI(Exception):
    """
    Raised when Tarantool returns some data (IPROTO_OK),
    but it's neither `return value` nor `return nil, err`.

    The actual returned data is contained in `self.args`.
    """

    ...

class ProcessDead(Exception):
    ...



def is_caused_by_timeout(e: BaseException) -> bool:
    """
    Check if there's a `TimeoutError` somewhere in the cause-chain of this exception.
    """
    if isinstance(e, TimeoutError):
        return True

    if not e.__context__:
        return False

    return is_caused_by_timeout(e.__context__)


def normalize_net_box_result(func):
    """
    Convert lua-style responses to be more python-like.
    This also fixes some of the connector API inconveniences.
    """

    def inner(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except DatabaseError as exc:
            if is_caused_by_timeout(exc):
                raise TimeoutError from exc

            if getattr(exc, "errno", None):
                # Error handling in Tarantool connector is awful.
                # It wraps NetworkError in DatabaseError.
                # We, instead, convert it to a native OSError.
                strerror = os.strerror(exc.errno)
                raise OSError(exc.errno, strerror) from exc

            source_location = None
            if exc.extra_info:
                source_location = f"{exc.extra_info.file}:{exc.extra_info.line}"

            match exc.args:
                case (int(code), arg):
                    # Error handling in Tarantool connector is awful.
                    # It returns error codes as raw numbers.
                    # Here we format them for easier use in pytest assertions.
                    error_info = tnt_strerror(code)
                    match error_info:
                        case (str(error_type), _):
                            raise TarantoolError(
                                error_type, arg, source_location
                            ) from exc
                        case "UNDEFINED":
                            raise TarantoolError(code, arg, source_location) from exc
                        case _:
                            raise RuntimeError("unreachable")
                case _:
                    raise exc from exc

        # This is special case for Connection.__init__
        if result is None:
            return

        match result.data:
            case []:
                return None
            case [x] | [x, None]:
                return x
            case [None, err]:
                raise ReturnError(err)
            case [*args]:
                raise MalformedAPI(*args)
            case _:
                raise RuntimeError("unreachable")

    return inner


class Connection(tarantool.Connection):
    @normalize_net_box_result
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @normalize_net_box_result
    def call(self, func_name, *args, on_push=None, on_push_ctx=None):
        return super().call(func_name, *args, on_push=on_push, on_push_ctx=on_push_ctx)

    @normalize_net_box_result
    def eval(self, expr, *args, on_push=None, on_push_ctx=None):
        return super().eval(expr, *args, on_push=on_push, on_push_ctx=on_push_ctx)

    def sql(self, sql: str, *params, options=None, sudo=False) -> dict[str, list]:
        """Run SQL query and return result"""
        if sudo:
            old_euid = self.eval(
                """
                local before = box.session.euid()
                box.session.su('admin')
                return before
                """
            )

        options = options or {}
        try:
            result = self.call(
                ".proc_sql_dispatch",
                sql,
                params,
                options.get("query_id"),
                options.get("traceable"),
            )[0]
        finally:
            if sudo:
                self.eval("box.session.su(...)", old_euid)

        return result


class TarantoolInstance:

    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
    ) -> None:
        self.connection: Optional[Connection] = None

        self.user = user
        self.host = host
        self.port = port
        self.password = password

    def start(self) -> None:
        self.connection = Connection(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
        )
        self.connection.connect()

    def eval(self, expr: str, *args: Any, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.eval(expr, *args, on_push=on_push, on_push_ctx=on_push_ctx)

    def update(self, space_name: str, key: Any, op_list: Any, *, index: int = 0, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.update(space_name, key, op_list, index=index, on_push=on_push, on_push_ctx=on_push_ctx)

    def delete(self, space_name: str, key: Any , *, index: int = 0, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.delete(space_name, key, index=index, on_push=on_push, on_push_ctx=on_push_ctx)

    def insert(self, space_name: str, values: Any, *, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.insert(space_name, values, on_push=on_push, on_push_ctx=on_push_ctx)

    def replace(self, space_name: str, values: Any, *, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.replace(space_name, values, on_push=on_push, on_push_ctx=on_push_ctx)

    def upsert(self, space_name: str, values: Any, op_list: Any, *, index: int = 0, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.upsert(space_name, values, op_list, index=index, on_push=on_push, on_push_ctx=on_push_ctx)

    def select(self, space_name: str, key: Any = None, *, offset: int = 0, limit: int = 0xffffffff, index: int = 0, iterator: Any = None, on_push: Any = None, on_push_ctx: Any = None) -> Any:
        return self.connection.select(space_name, key, offset=offset, limit=limit, index=index, iterator=iterator, on_push=on_push, on_push_ctx=on_push_ctx)

    def stop(self) -> None:
        self.connection.close()


class UroborosInstance:
    def __init__(
            self,
            host: str,
            port: int,
            user: str,
            password: str,
            is_master: bool,
            data_dir: str,
            master_port: int,
            http_port: int,
        ) -> None:
        self.connection: Optional[Connection] = None
        self.process: subprocess.Popen | None = None

        self.user = user
        self.host = host
        self.port = port
        self.http_port = http_port
        self.data_dir = data_dir
        self.password = password
        self.is_master = is_master
        self.master_port = master_port

        self.plugin_dir = "./local-dev-clusters/plugins"
        self.init_replication_factor = 2
        self.service_password_file = "password.txt"


    @property
    def command(self) -> list[str]:
        args = [
            PICO_BINARY_PATH,
            "run",
            f"--plugin-dir={self.plugin_dir}",
            f"--listen={self.host}:{self.port}",
            f"--peer={self.host}:{self.master_port}",
            f"--service-password-file={self.service_password_file}",
            f"--data-dir={self.data_dir}",
            f"--http-listen={self.host}:{self.http_port}",
        ]
        if self.is_master:
            args.append(f"--init-replication-factor={self.init_replication_factor}")
        return args

    def call(self, func_name: str, *args: Any, on_push: Any=None, on_push_ctx: Any=None) -> Any:
        return self.connection.call(func_name, *args, on_push=on_push, on_push_ctx=on_push_ctx)

    def eval(self, func_name: str, *args: Any, on_push: Any=None, on_push_ctx: Any=None) -> Any:
        return self.connection.eval(func_name, *args, on_push=on_push, on_push_ctx=on_push_ctx)


    def wait_online(
        self, timeout: int | float = 30, rps: int | float = 5, expected_incarnation: Any = None
    ) -> None:

        if self.process is None:
            raise ProcessDead("process was not started")


        def fetch_current_state() -> Tuple[str, int]:
            self.check_process_alive()

            myself = self.call(".proc_instance_info")
            assert isinstance(myself, dict)

            assert isinstance(myself["current_state"], dict)
            return (
                myself["current_state"]["variant"],
                myself["current_state"]["incarnation"],
            )

        start = time.monotonic()
        deadline = start + timeout
        next_retry = start
        last_state = None
        while True:
            now = time.monotonic()
            assert now < deadline, "timeout"

            if now < next_retry:
                time.sleep(next_retry - now)
            next_retry = time.monotonic() + 1 / rps

            try:
                state = fetch_current_state()
                if state != last_state:
                    last_state = state
                    deadline = time.monotonic() + timeout
                variant, incarnation = state
                assert variant == "Online"
                if expected_incarnation is not None:
                    assert incarnation == expected_incarnation
                break

            except ProcessDead as e:
                raise e from e
            except Exception as e:
                match e.args:
                    case (ErrorCode.Loading, _):
                        if time.monotonic() > start + timeout * 6:
                            raise e from e
                    case _:
                        if time.monotonic() > deadline:
                            raise e from e

        eprint(f"{self} is online")

    def check_process_alive(self):
        if self.process is None:
            raise ProcessDead("process was not started")
        try:
            exit_code = self.process.wait(timeout=0.1)
        except subprocess.TimeoutExpired:
            pass
        else:
            raise ProcessDead(f"process exited unexpectedly, {exit_code=}")

    def _process_output(self, src: Any, out: io.TextIOWrapper) -> None:
        id = f":{self.port}"
        prefix = f"{id:<3} | "

        prefix_bytes = prefix.encode("utf-8")

        for line in iter(src.readline, b""):
            with OUT_LOCK:
                out.buffer.write(prefix_bytes)
                out.buffer.write(line)
                out.flush()

        src.close()

    def start(self) -> None:
        if self.process:
            return
        try:
            os.makedirs(self.data_dir)
        except Exception:
            ...
        env = {}
        if not os.environ.get("PICODATA_LOG_LEVEL") and "PICODATA_LOG_LEVEL" not in env:
            env.update(PICODATA_LOG_LEVEL="warn")
        if os.environ.get("RUST_BACKTRACE") is not None:
            env.update(RUST_BACKTRACE=str(os.environ.get("RUST_BACKTRACE")))
        if os.getenv("NOLOG"):
            out = subprocess.DEVNULL
        else:
            out = subprocess.PIPE
        self.process = subprocess.Popen(
            self.command,
            env=env or None,
            stdin=subprocess.DEVNULL,
            stdout=out,
            stderr=out,
            start_new_session=True,
        )
        assert os.getpgid(self.process.pid) == self.process.pid
        if out == subprocess.DEVNULL:
            return
        for src, out in [
            (self.process.stdout, sys.stdout),
            (self.process.stderr, sys.stderr),
        ]:
            threading.Thread(
                target=self._process_output,
                args=(src, out),
                daemon=True,
            ).start()

        tries = 0

        while tries < 20:
            tries += 1
            try:
                self.connection = Connection(
                    host=self.host,
                    port=self.port,
                    user=self.user,
                    password=self.password,
                )
                self.connection.connect()
                break
            except Exception:
                time.sleep(_1_SEC * 0.5)
                continue

    def stop(self) -> None:
        self.connection.close()
        if self.process is None:
            return
        pid = self.process.pid
        with suppress(ProcessLookupError, PermissionError):
            os.killpg(pid, signal.SIGKILL)
            with suppress(ChildProcessError):
                os.waitpid(pid, 0)
        self.process = None
        try:
            shutil.rmtree(self.data_dir)
        except Exception:
            ...


class UroborosCluster:
    def __init__(self, master: UroborosInstance, slaves: list[UroborosInstance]) -> None:
        self.master = master
        self.slaves = slaves

    def start(self) -> None:
        self.master.start()
        for slave in self.slaves:
            slave.start()
        self.master.wait_online()
        for slave in self.slaves:
            slave.wait_online()
        self.master.eval("""return pico.install_plugin("uroboros", "0.2.9")""")
        self.master.eval("""return pico.migration_up("uroboros", "0.2.9")""")
        self.master.eval("""return pico.service_append_tier("uroboros", "0.2.9", "uroboros", "default")""")
        self.master.eval("""return pico.enable_plugin("uroboros", "0.2.9")""")
        # TODO: track some state on uroboros to not sleep
        time.sleep(_1_SEC * 25)

    def stop(self) -> None:
        for slave in self.slaves:
            slave.stop()
        self.master.stop()

    def get_state(self) -> dict:
        result = {}
        state = self.master.eval("""
        local res = {}
        for _, v in box.space.state:pairs() do table.insert(res, v:tomap({names_only=true})) end
        return res
        """)
        for i in state:
            result[i['cluster_uuid']] = i['lsn']
        return result


class TarantoolCluster:

    def __init__(self, instances: list[TarantoolInstance]) -> None:
        self.spaces: dict[str, list[dict[str,str]]] = {}
        self.indexes: dict[str, set[str]] = defaultdict(set)

        self.instances = instances

    def start(self) -> None:
        for instance in self.instances:
            instance.start()
        self.fetch_spaces()
        self.fetch_indexes()

    def stop(self) -> None:
        for instance in self.instances:
            instance.stop()

    def fetch_spaces(self) -> None:
        space_names_lua = """
        local result = {}
        for _, space in box.space._vspace:pairs() do
            local space_name = space[3]
            local format = box.space[space_name]:format()
            table.insert(result, {space_name, format})
        end
        return result
        """
        spaces_by_instance = []
        for instance in self.instances:
            spaces_by_instance.append({space[0]: space[1] for space in instance.eval(space_names_lua)})
        current_spaces = spaces_by_instance[0]
        self.spaces = current_spaces

    def fetch_indexes(self) -> None:
        indexes_lua = """
        local result = {}
        for _, index in box.space._vindex:pairs() do
            local space_id = index[1]
            local index_id = index[2]
            local index_name = index[3]
            local index_parts = index[6]
            local space_name = box.space._vspace:get(space_id)[3]
            if not result[space_name] then
                result[space_name] = {}
            end
            table.insert(result[space_name], {
                index_id = index_id,
                index_name = index_name,
                index_parts = index_parts
            })
        end
        return result
        """
        indexes_by_instance = []
        for instance in self.instances:
            map = defaultdict(set)
            indexes = instance.eval(indexes_lua)
            for space, indexes in indexes.items():
                for index in indexes:
                    map[space].add(index["index_name"])
            indexes_by_instance.append(map)
        current_indexes = indexes_by_instance[0]
        self.indexes = current_indexes

    def space_format(self, space_name: str) -> Optional[list[dict[str, str]]]:
        return self.spaces.get(space_name)

    def space_indexes(self, space_name: str) -> set[str]:
        return self.indexes[space_name]

    def create_space(self, space_name: str, fields: list[dict[str, str]], is_sync: bool = False, is_sharded: bool = True) -> None:
        create_space_lua = f"""
        box.schema.space.create('{space_name}', {{engine='memtx', is_sync={str(is_sync).lower()}}})
        box.space.{space_name}:format({format_format(fields)})
        box.space.{space_name}:create_index('primary', {{parts = {{1, 'unsigned'}}}})
        """
        if is_sharded:
            create_space_lua = create_space_lua + f"""
            box.space.{space_name}:create_index('bucket_id', {{parts = {{'bucket_id'}}, unique = false}})
            """
        for instance in self.instances:
            instance.eval(create_space_lua)
        self.spaces[space_name] = fields
        self.indexes[space_name].add("primary")
        if is_sharded:
            self.indexes[space_name].add("bucket_id")

    def drop_space(self, space_name: str) -> None:
        assert space_name in self.spaces, "space does not exist"
        drop_space_lua = f"box.space.{space_name}:drop()"
        for instance in self.instances:
            instance.eval(drop_space_lua)
        del self.spaces[space_name]
        del self.indexes[space_name]

    def add_field(self, space_name: str, field_name: str, field_type: str) -> None:
        add_field_lua = f"""
        local space = box.space.{space_name}
        local format = space:format()
        table.insert(format, {{name = '{field_name}', type = '{field_type}'}})
        space:format(format)
        """
        for instance in self.instances:
            instance.eval(add_field_lua)

    def remove_field(self, space_name: str, field_name: str) -> None:
        remove_field_lua = f"""
        local space = box.space.{space_name}
        local format = space:format()
        for i, field in ipairs(format) do
            if field.name == '{field_name}' then
                table.remove(format, i)
                break
            end
        end
        space:format(format)
        """
        for instance in self.instances:
            instance.eval(remove_field_lua)

    def create_index(self, space_name: str, index_name: str, part: tuple[str, str], unique: bool = True) -> None:
        i = "{" + f"{part[0]}, '{part[1]}'" + "}"
        create_index_lua = f"""
        box.space.{space_name}:create_index('{index_name}', {{
            parts = {i},
            unique = {str(unique).lower()}
        }})
        """
        for instance in self.instances:
            instance.eval(create_index_lua)
        self.indexes[space_name].add(index_name)

    def drop_index(self, space_name: str, index_name: str) -> None:
        drop_index_lua = f"box.space.{space_name}.index.{index_name}:drop()"
        for instance in self.instances:
            instance.eval(drop_index_lua)
        self.indexes[space_name].remove(index_name)

    def insert(self, space_name: str, key: Any, values: tuple[Any, ...]) -> None:
        idx = hash(key) % len(self.instances)
        self.instances[idx].insert(space_name, values)

    def update(self, space_name: str, key: Any, operations: list[tuple[Any, ...]]) -> None:
        idx = hash(key) % len(self.instances)
        self.instances[idx].update(space_name, key, operations)

    def delete(self, space_name: str, key: Any) -> None:
        idx = hash(key) % len(self.instances)
        self.instances[idx].delete(space_name, key)

    def replace(self, space_name: str, key: Any, values: tuple[Any, ...]) -> None:
        idx = hash(key) % len(self.instances)
        self.instances[idx].replace(space_name, values)

    def upsert(self, space_name: str, key: Any, values: tuple[Any, ...], operations: list[tuple[Any, ...]]) -> None:
        idx = hash(key) % len(self.instances)
        self.instances[idx].upsert(space_name, values, operations)

    def select(self, space_name: str, key: Any) -> list[tuple[Any, ...]]:
        result = []
        for instance in self.instances:
            selected = instance.select(space_name, key)
            result.extend([tuple(i) for i in selected.data])
        return result

    def get_lsns(self) -> dict:
        result = {}
        info_lua = f"""
        return {{ cluster_uuid = box.info.cluster.uuid, lsn = box.info.lsn }}
        """
        for instance in self.instances:
            info = instance.eval(info_lua)
            result[info['cluster_uuid']] = info['lsn']
        return result

    def fill_snapshots(self) -> None:
        url = 'http://localhost:9001/customer'
        for i in range(10):
            body = { 'customer_id': i + 1, 'name': 'Абырвалг' }
            requests.post(url, json = body)
        for instance in self.instances:
            instance.eval("box.snapshot()")


@pytest.fixture(scope="session", autouse=True)
def setup_clusters() -> Generator[None, None, None]:
    if TEST_RUN_CLUSTERS == "docker":
        subprocess.run(["make", "run_docker_clusters"], check=True)
    elif TEST_RUN_CLUSTERS == "cartridge":
        subprocess.run(["make", "run_cartridge_clusters"], check=True)
    yield
    if TEST_RUN_CLUSTERS == "docker":
        subprocess.run(["make", "stop_docker_clusters"], check=True)
    elif TEST_RUN_CLUSTERS == "cartridge":
        subprocess.run(["make", "stop_cartridge_clusters"], check=True)

@pytest.fixture(scope="session", autouse=True)
def uroboros_cluster() ->  Generator[UroborosCluster, None, None]:
    master = UroborosInstance(
        host=BASE_HOST,
        port=4401,
        user=PICO_SERVICE_USER,
        password="uroboros",
        is_master=True,
        data_dir="./local-dev-clusters/one",
        master_port=4401,
        http_port=9091,
    )
    slave = UroborosInstance(
        host=BASE_HOST,
        port=4402,
        user=PICO_SERVICE_USER,
        password="uroboros",
        is_master=False,
        data_dir="./local-dev-clusters/two",
        master_port=4401,
        http_port=9092,
    )

    cluster = UroborosCluster(master=master, slaves=[slave])
    cluster.start()
    yield cluster
    cluster.stop()

@pytest.fixture(scope="session")
def producer_cluster() -> Generator[TarantoolCluster, None, None]:
    cluster = TarantoolCluster(
        [
        TarantoolInstance(
            host=BASE_HOST,
            port=3312,
            user=BASE_USER,
            password=BASE_PASSWORD
        ),
        TarantoolInstance(
            host=BASE_HOST,
            port=3314,
            user=BASE_USER,
            password=BASE_PASSWORD
        ),
        TarantoolInstance(
            host=BASE_HOST,
            port=3316,
            user=BASE_USER,
            password=BASE_PASSWORD
        )
    ]
    )
    cluster.start()
    cluster.fill_snapshots()
    yield cluster
    cluster.stop()


@pytest.fixture(scope="session")
def consumer_cluster() -> Generator[TarantoolCluster, None, None]:
    cluster = TarantoolCluster(
        instances=[
        TarantoolInstance(
            host=BASE_HOST,
            port=3322,
            user=BASE_USER,
            password=BASE_PASSWORD
        ),
        TarantoolInstance(
            host=BASE_HOST,
            port=3324,
            user=BASE_USER,
            password=BASE_PASSWORD
        ),
    ]
    )
    cluster.start()
    yield cluster
    cluster.stop()
