


from time import sleep
from conftest import TarantoolCluster, UroborosCluster

_1_SEC = 1

# test for correct snapshot replication - snapshot is created on cluster start
def test_snapshot_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    customer_space = "customer"
    customer_format = [   
        { 'name': 'customer_id', 'type': 'unsigned' },
        { 'name': 'bucket_id',   'type': 'unsigned' },
        { 'name': 'name',        'type': 'string' },
    ]
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(customer_space) == customer_format
    
    produced_customer_tuples = producer_cluster.select(customer_space, [])
    produced_customers = {}
    for t in produced_customer_tuples:
        produced_customers[t[0]] = t
    consumed_customer_tuples = consumer_cluster.select(customer_space, [])
    consumed_customers = {}
    for t in consumed_customer_tuples:
        consumed_customers[t[0]] = t
    assert len(produced_customers) > 0
    assert produced_customers == consumed_customers    


# test for correct uroboros work if there is a COMMIT message but limbo is empty
def test_skip_non_replicated_transaction(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster, uroboros_cluster: UroborosCluster) -> None:
    space_name = "test_trans_non_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format, is_sync=True, is_sharded=False)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == None
    
    produced_tuples = []
    for key in range(1, 3):
        t = (key, f"name_{key}", key + 10)
        producer_cluster.insert(space_name, key, t)
        produced_tuples.append(t)
        
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == None

    readed_producer_tuples = producer_cluster.select(space_name, [])
    assert produced_tuples == readed_producer_tuples
    
    sleep(_1_SEC * 5)
    
    is_uroboros_alive = False
    exception = None
    try:
        uroboros_cluster.master.wait_online(3)
        for instance in uroboros_cluster.slaves:
            instance.wait_online(3)
        is_uroboros_alive = True
    except Exception as e:
        is_uroboros_alive = False
        exception = e
    
    assert is_uroboros_alive == True
    assert exception is None


def test_lsn_for_skipped(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster, uroboros_cluster: UroborosCluster) -> None:
    space_name = "test_trans_non_replicated_lsn"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format, is_sharded=False)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == None
    
    produced_tuples = []
    for key in range(1, 3):
        t = (key, f"name_{key}", key + 10)
        producer_cluster.insert(space_name, key, t)
        produced_tuples.append(t)
        
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == None

    readed_producer_tuples = producer_cluster.select(space_name, [])
    assert produced_tuples == readed_producer_tuples
    
    sleep(_1_SEC * 5)

    is_uroboros_alive = False
    exception = None
    try:
        uroboros_cluster.master.wait_online(3)
        for instance in uroboros_cluster.slaves:
            instance.wait_online(3)
        is_uroboros_alive = True
    except Exception as e:
        is_uroboros_alive = False
        exception = e
    
    assert is_uroboros_alive == True
    assert exception is None

    producer_lsns = producer_cluster.get_lsns()
    state = uroboros_cluster.get_state()
    for cluster_uuid in producer_lsns.keys():
        assert not state[cluster_uuid] is None
        assert state[cluster_uuid] == producer_lsns[cluster_uuid]


def test_insert_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    space_name = "test_insert_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
        {'name': 'bucket_id', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == space_format
    
    produced_tuples = []
    for key in range(1, 101):
        t = (key, f"name_{key}", key + 10, key)
        producer_cluster.insert(space_name, key, t)
        produced_tuples.append(t)
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples
    

def test_update_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    space_name = "test_update_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
        {'name': 'bucket_id', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == space_format
    
    produced_tuples = []
    for key in range(1, 101):
        t = (key, f"name_{key}", key + 10, key)
        producer_cluster.insert(space_name, key, t)
        produced_tuples.append(t)
    
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples
    
    for i in range(len(produced_tuples)):
        key = produced_tuples[i][0]
        new_name = f"new_name_{key}"
        new_age = i * 100
        operations = [
            ("=", 1, new_name),
            ("=", 2, new_age),
        ]
        producer_cluster.update(space_name, key, operations)
        produced_tuples[i] = (key, new_name, new_age, produced_tuples[i][3])
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples
        
    
    
def test_replace_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    space_name = "test_replace_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
        {'name': 'bucket_id', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == space_format
    
    produced_tuples = []
    for key in range(1, 101):
        t = (key, f"name_{key}", key + 10, key)
        producer_cluster.replace(space_name, key, t)
        produced_tuples.append(t)
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples


def test_upsert_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    space_name = "test_upsert_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
        {'name': 'bucket_id', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == space_format
    
    produced_tuples = []
    for key in range(1, 101):
        t = (key, f"name_{key}", key + 10, key)
        producer_cluster.replace(space_name, key, t)
        produced_tuples.append(t)
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples

    updated_tuples = []
    for i in range(len(produced_tuples)):
        key = produced_tuples[i][0]
        new_name = f"new_name_{key}"
        new_age = i * 100
        new_tuple = (key, new_name, new_age, key)
        operations = [
            ("=", 1, new_name),
            ("=", 2, new_age),
        ]
        producer_cluster.upsert(space_name, key, new_tuple, operations)
        updated_tuples.append(new_tuple)
    for key in range(201, 220):
        t = (key, f"name_{key}", key + 10, key)
        operations = [
            ("=", 1, f"name_{key}"),
            ("=", 2, key + 10),
        ]
        producer_cluster.upsert(space_name, key, t, operations)
        updated_tuples.append(t)
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert updated_tuples == consumed_tuples


def test_delete_replicated(producer_cluster: TarantoolCluster, consumer_cluster: TarantoolCluster) -> None:
    space_name = "test_delete_replicated"
    space_format = [   
        {'name': 'id', 'type': 'unsigned'},
        {'name': 'name', 'type': 'string'},
        {'name': 'age', 'type': 'unsigned'},
        {'name': 'bucket_id', 'type': 'unsigned'},
    ]
    producer_cluster.create_space(space_name, space_format)
    sleep(_1_SEC)
    consumer_cluster.fetch_spaces()
    assert consumer_cluster.space_format(space_name) == space_format
    
    produced_tuples = []
    for key in range(1, 101):
        t = (key, f"name_{key}", key + 10, key)
        producer_cluster.insert(space_name, key, t)
        produced_tuples.append(t)
        
    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert produced_tuples == consumed_tuples
    
    for key in range(1, 101):
        producer_cluster.delete(space_name, key)
        
    for t in produced_tuples:
        producer_cluster.delete(space_name, t[0])

    sleep(_1_SEC * 7)
    
    consumed_tuples = consumer_cluster.select(space_name, [])
    
    assert len(consumed_tuples) == 0
