#![allow(static_mut_refs)]
#![allow(clippy::unsafe_derive_deserialize)]

use picoplugin::system::tarantool::tlua;

pub static mut PASSWORD: String = String::new();

pub const RPC_REPLICATOR_ADD_ROUTE: &str = "/replicator_add";
pub const RPC_COORDINATOR_DDL_ROUTE: &str = "/coordinator_ddl";
pub const RPC_REPLICATOR_REMOVE_ROUTE: &str = "/replicator_remove";
pub const RPC_REPLICATOR_CONFIGURE_ROUTE: &str = "/replicator_configure";
pub const RPC_REPLICATOR_WRITERS_ROUTE: &str = "/replicator_writers";

pub const SIDECAR_APPLY_DDL_PROC: &str = "liburoboros_apply_ddl";

pub const SPACE_SPACE_ID: u32 = 280;
pub const INDEX_SPACE_ID: u32 = 288;

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct CredentialsConfig {
    pub id: String,
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct TarantoolWriterConfig {
    pub vshard: VShardGroupConfig,
    pub buckets_per_writer: u64,
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct KafkaWriterConfig {}

#[derive(
    Clone, Debug, tlua::LuaRead, tlua::Push, tlua::PushInto, serde::Serialize, serde::Deserialize,
)]
pub enum WriterConfig {
    Kafka(KafkaWriterConfig),
    Tarantool(TarantoolWriterConfig),
}

impl Default for WriterConfig {
    fn default() -> Self {
        Self::Tarantool(TarantoolWriterConfig::default())
    }
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct JobConfig {
    pub leader_id: String,
    pub replicaset_id: String,
    pub replicaset_alias: Option<String>,
    pub vshard_group: String,
    pub reader: ReaderConfig,
    pub disabled_spaces: Vec<String>,
}

#[derive(
    Clone,
    Debug,
    Default,
    PartialEq,
    Eq,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct SpaceConfig {
    pub space_id: u32,
    pub space_name: String,
    pub bucket_id_index: usize,
}

#[derive(
    Clone,
    Debug,
    Default,
    PartialEq,
    Eq,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct ServerConfig {
    pub uuid: String,
    pub uri: String,
    pub is_active_master: bool,
}

impl ServerConfig {
    /// Parses server uri into (username, password, host, port) parts.
    ///
    /// # Panics
    ///
    /// - In case of melformed uri.
    ///
    /// # Returns
    ///
    /// All parts of uri.
    #[must_use]
    pub fn parse_uri(&self) -> (&str, &str, &str, u16) {
        let mut split = self.uri.split(':');
        let host = split.next().expect("host should exist in addr");
        let port = split
            .next()
            .expect("port should exist in addr")
            .parse::<u16>()
            .expect("port should be valid u16");
        ("uroboros", unsafe { &PASSWORD }, host, port)
    }
}

#[derive(
    Clone,
    Debug,
    Default,
    PartialEq,
    Eq,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct ReplicasetConfig {
    pub vshard_group: String,
    pub uuid: String,
    pub alias: Option<String>,
    pub spaces: Vec<SpaceConfig>,
    pub servers: Vec<ServerConfig>,
    pub is_router: bool,
    pub is_storage: bool,
}

impl ReplicasetConfig {
    /// Get a master for replicaset.
    ///
    /// # Returns
    ///
    /// A reference to a `ServerConfig`.
    #[must_use]
    pub fn master(&self) -> &ServerConfig {
        for replica in &self.servers {
            if !replica.is_active_master {
                continue;
            }
            return replica;
        }
        unreachable!("there must be a master in replicaset")
    }
}

#[derive(
    Clone,
    Debug,
    Default,
    PartialEq,
    Eq,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct VShardGroupConfig {
    pub name: String,
    pub bucket_count: u64,
    pub replicasets: Vec<ReplicasetConfig>,
}

impl VShardGroupConfig {
    /// Get replicaset by replica addr.
    ///
    /// # Returns
    ///
    /// A reference to a `ReplicasetConfig`.
    #[must_use]
    pub fn get_shard_by_replica(&self, addr: &str) -> &ReplicasetConfig {
        for shard in &self.replicasets {
            for replica in &shard.servers {
                if replica.uri == addr {
                    return shard;
                }
            }
        }
        unreachable!("there must be a shard for replica: {addr}")
    }
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::PushInto,
    tlua::Push,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct ReaderConfig {
    pub addr: String,
    pub username: String,
    pub password: String,
    pub vshard_group: String,
    pub replicaset_id: String,
    pub replica_uuid: String,
}

impl From<ReplicasetConfig> for ReaderConfig {
    fn from(value: ReplicasetConfig) -> Self {
        let master = value.master();
        let (username, password, host, port) = master.parse_uri();
        Self {
            addr: format!("{host}:{port}"),
            username: username.to_string(),
            password: password.to_string(),
            vshard_group: value.vshard_group.clone(),
            replicaset_id: value.uuid.clone(),
            replica_uuid: master.uuid.clone(),
        }
    }
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct TarantoolConfig {
    pub user_url: String,
    pub topology_url: String,
    pub space_info_url: String,
}

#[derive(
    Clone,
    Debug,
    Default,
    tlua::LuaRead,
    tlua::Push,
    tlua::PushInto,
    serde::Serialize,
    serde::Deserialize,
)]
pub struct KafkaConfig {}

#[derive(
    Clone, Debug, tlua::LuaRead, tlua::Push, tlua::PushInto, serde::Serialize, serde::Deserialize,
)]
#[serde(tag = "type", content = "attributes")]
#[serde(rename_all = "snake_case")]
pub enum ConsumerConfig {
    Kafka(KafkaConfig),
    Tarantool(TarantoolConfig),
}

impl Default for ConsumerConfig {
    fn default() -> Self {
        Self::Tarantool(TarantoolConfig::default())
    }
}

#[derive(
    Clone, Debug, Default, tlua::LuaRead, tlua::Push, serde::Serialize, serde::Deserialize,
)]
pub struct Config {
    /// Password which is used for role
    pub password: String,
    /// Spaces which will not be replicated.
    pub disabled_spaces: Vec<String>,
    /// Groups which will be replicated.
    pub enabled_groups: Vec<String>,
    /// Address of cluster which produces data.
    pub producer: TarantoolConfig,
    /// Address of right cluster which is used to get topology.
    pub consumer: ConsumerConfig,
    /// Number of buckets per writer
    pub buckets_per_writer: u64,
}

impl VShardGroupConfig {
    /// Convert config to lua table so it can be simply used within lua.
    ///
    /// # Returns
    ///
    /// Message pack encoded into lua string.
    #[must_use]
    pub fn as_table(self) -> tlua::AnyLuaString {
        let mut map = Vec::with_capacity(self.replicasets.len());
        for sharding in self.replicasets {
            if !sharding.is_storage {
                // if replicaset hasn't role vshard-storage then skip it
                continue;
            }
            let mut semap = Vec::with_capacity(sharding.servers.len());
            for replica in sharding.servers {
                let uri = unsafe { format!("uroboros:{PASSWORD}@{}", replica.uri) };
                semap.push((
                    rmpv::Value::String(replica.uuid.into()),
                    rmpv::Value::Map(vec![
                        (
                            rmpv::Value::String(String::from("uri").into()),
                            rmpv::Value::String(uri.into()),
                        ),
                        (
                            rmpv::Value::String(String::from("master").into()),
                            rmpv::Value::Boolean(replica.is_active_master),
                        ),
                    ]),
                ));
            }
            map.push((
                rmpv::Value::String(sharding.uuid.into()),
                rmpv::Value::Map(vec![(
                    rmpv::Value::String(String::from("replicas").into()),
                    rmpv::Value::Map(semap),
                )]),
            ));
        }

        let table = rmpv::Value::Map(vec![
            (
                rmpv::Value::String(String::from("bucket_count").into()),
                rmpv::Value::Integer(self.bucket_count.into()),
            ),
            (
                rmpv::Value::String(String::from("discovery_mode").into()),
                rmpv::Value::String(String::from("off").into()),
            ),
            (
                rmpv::Value::String(String::from("sharding").into()),
                rmpv::Value::Map(map),
            ),
        ]);

        tlua::AnyLuaString(utils::value_to_vec(&table))
    }
}
