use std::{collections, rc, time};

use picoplugin::system::tarantool::{fiber, net_box};

use crate::{error, types, SidecarDDLType};

pub struct Fiber {
    vshard_group_name: String,

    client: fibreq::Client,

    producer_instances: types::ProducerInstances,
    consumer_connections: types::ConsumerConnections,

    close_chan: fiber::Channel<()>,
    ddl_chan: fiber::Channel<types::DDL>,
    jobs_chan: fiber::Channel<types::Job>,
    configure_chan: fiber::Channel<()>,

    ddl_cond: rc::Rc<fiber::Cond>,

    current_ddl: Option<types::DDL>,

    producer_cfg: config::TarantoolConfig,
}

impl Fiber {
    pub fn new(
        vshard_group_name: String,
        producer_instances: types::ProducerInstances,
        consumer_connections: types::ConsumerConnections,
        jobs_chan: fiber::Channel<types::Job>,
        configure_chan: fiber::Channel<()>,
        producer_cfg: config::TarantoolConfig,
    ) -> Self {
        let client = fibreq::ClientBuilder::new().build();
        let ddl_cond = rc::Rc::new(fiber::Cond::new());
        let ddl_chan = fiber::Channel::new(1);
        let close_chan = fiber::Channel::new(1);
        Self {
            client,
            vshard_group_name,
            producer_instances,
            consumer_connections,
            close_chan,
            ddl_chan,
            jobs_chan,
            ddl_cond,
            producer_cfg,
            configure_chan,
            current_ddl: None,
        }
    }

    pub fn start(mut self) -> Handle {
        let vshard_group_name = self.vshard_group_name.clone();
        log::warn!("start applier fiber for vshard group: {vshard_group_name}",);
        let ddl_cond = self.ddl_cond.clone();
        let close_chan = self.close_chan.clone();
        let ddl_chan = self.ddl_chan.clone();
        let handle = fiber::Builder::new()
            .func(move || loop {
                if self.close_chan.try_recv().is_ok() {
                    return;
                }
                if self.ddl_chan.is_empty() {
                    fiber::sleep(time::Duration::from_millis(100));
                    continue;
                }
                self.collect();
                if let Ok(ddls) = self.get_sharded_ddl() {
                    self.apply(&ddls);
                }
                self.release();
            })
            .name(format!("uroboros_applier_{vshard_group_name}"))
            .start()
            .expect("starting applier fiber should not fail");
        Handle {
            close_chan,
            fiber: Some(handle),
            ddl_chan,
            ddl_cond,
        }
    }

    fn collect(&mut self) {
        let mut count = 0;
        let producer_guard = self.producer_instances.lock();
        let connections_total = producer_guard
            .get(&self.vshard_group_name)
            .expect("vshard group connections should exist")
            .to_owned();
        drop(producer_guard);
        log::info!("will collect ddl from, want ddl from {connections_total} instances");
        while count < connections_total {
            log::info!("wait for ddl {}/{}", count, connections_total);
            let d = self
                .ddl_chan
                .recv()
                .expect("recv from ddl from chan should not fail");
            log::info!("received ddl {:?}", &d);
            if self.current_ddl.is_none() {
                self.current_ddl = Some(d);
            }
            count += 1;
            log::info!("received ddl {}/{}", count, connections_total);
        }
        log::info!("collected ddl from {connections_total} instances");
    }

    fn get_space_name(&mut self, ddl: &types::DDL) -> Option<String> {
        if let Some(v) = &ddl.ddl.target_space_name {
            return Some(v.clone());
        }

        let data_vec = if ddl.ddl.key.is_empty() {
            ddl.ddl.data.clone()
        } else {
            ddl.ddl.key.clone()
        };
        let space_id_from_ddl: u32 = u32::try_from(
            utils::vec_to_value(data_vec)
                .as_array()
                .expect("value should be a tuple") // space id is 1st value in tuple
                .first()
                .expect("tuple should be an array")
                .as_u64()
                .expect("first element of tuple should be u64"),
        )
        .expect("space_id should be u32");

        let topology = utils::tx_retry!(self.fetch_topology(), delay: 100, tries: 10, backoff: 2.0)
            .expect("failed to update topology on applier release");
        topology
            .into_iter()
            .filter(|x| x.name == ddl.vshard_group)
            .flat_map(|x| x.replicasets)
            .filter(|x| x.uuid == ddl.replicaset_id)
            .flat_map(|x| x.spaces)
            .filter(|x| x.space_id == space_id_from_ddl)
            .map(|x| x.space_name)
            .nth(0)
    }

    fn get_sharded_ddl(&mut self) -> Result<Vec<types::DDL>, error::Error> {
        if self.current_ddl.is_none() {
            return Ok(vec![]);
        }
        let ddl = self.current_ddl.take().expect("ddl should not be none");
        // exclude all except 280 (_space) and 288 (_index)
        if ddl.ddl.space_id != config::SPACE_SPACE_ID && ddl.ddl.space_id != config::INDEX_SPACE_ID
        {
            return Ok(vec![]);
        }
        if let SidecarDDLType::Delete = ddl.ddl.kind {
            return Ok(vec![ddl]);
        }

        log::info!("get_sharded_ddl: ddl {:?}", ddl);
        let space_name = self.get_space_name(&ddl);
        if space_name.is_none() {
            return Ok(vec![]);
        }

        if let Some(space) = self.fetch_space_info(&ddl.vshard_group, &space_name.unwrap())? {
            let mut ddls = Vec::with_capacity(&space.indexes.len() + 1);
            ddls.push(types::DDL {
                vshard_group: ddl.vshard_group.clone(),
                replicaset_id: ddl.replicaset_id.clone(),
                ddl: crate::SidecarDDL {
                    space_id: config::SPACE_SPACE_ID,
                    key: ddl.ddl.key.clone(),
                    data: hex::decode(space.space).map_err(|err| {
                        log::error!("decode from hex error {err}");
                        error::Error::FetchSpaceInfo
                    })?,
                    kind: crate::SidecarDDLType::Replace,
                    target_space_name: Some(space.name.clone()),
                },
            });
            for idx in space.indexes {
                ddls.push(types::DDL {
                    vshard_group: ddl.vshard_group.clone(),
                    replicaset_id: ddl.replicaset_id.clone(),
                    ddl: crate::SidecarDDL {
                        space_id: config::INDEX_SPACE_ID,
                        key: ddl.ddl.key.clone(),
                        data: hex::decode(idx).map_err(|err| {
                            log::error!("decode from hex error {err}");
                            error::Error::FetchSpaceInfo
                        })?,
                        kind: crate::SidecarDDLType::Replace,
                        target_space_name: Some(space.name.clone()),
                    },
                });
            }
            Ok(ddls)
        } else {
            Ok(vec![])
        }
    }

    fn apply(&mut self, ddls: &[types::DDL]) {
        if ddls.is_empty() {
            return;
        }

        let consumer_guard = self.consumer_connections.lock();
        let connections = consumer_guard
            .get(&self.vshard_group_name)
            .expect("vshard group connections should exist");
        // Initialy set retry set to contain each replicaset within vshard group
        let mut retry = collections::HashSet::with_capacity(connections.len());
        let mut applied = collections::HashSet::with_capacity(connections.len());
        for replicaset_id in connections.keys() {
            retry.insert(replicaset_id.to_owned());
        }
        drop(consumer_guard);

        while !retry.is_empty() {
            log::info!("will apply ddls for repicasets: {retry:?}");
            // We recreate this lock on each iteration so the fiber which keeps connections
            // could recrete them in case of changed master/destroyed node
            let consumer_guard = self.consumer_connections.lock();
            let connections = consumer_guard
                .get(&self.vshard_group_name)
                .expect("vshard group connections should exist");
            'inner: for replicaset_id in connections.keys() {
                if applied.contains(replicaset_id) {
                    continue 'inner;
                }
                retry.insert(replicaset_id.to_owned());
            }
            retry.retain(|x| connections.contains_key(x));
            log::info!("cleared retry set to contain only actual replicasets which want to receive ddl: {retry:?}");
            'inner: for (replicaset_id, master) in connections {
                if !retry.contains(replicaset_id) {
                    continue 'inner;
                }
                log::info!("apply ddls for replicaset: {replicaset_id}");
                for ddl in ddls {
                    log::info!("apply ddl: {:?}", &ddl.ddl);
                    match master.call(
                        config::SIDECAR_APPLY_DDL_PROC,
                        &(&ddl.ddl,),
                        &net_box::Options::default(),
                    ) {
                        Ok(_) => {
                            retry.remove(replicaset_id);
                            applied.insert(replicaset_id.to_owned());
                        }
                        Err(e) => {
                            log::error!(
                                "failed to apply ddl for: {}, with: {:?}",
                                replicaset_id,
                                e
                            );
                        }
                    }
                    log::info!("applied ddl: {ddl:?} for repicasets: {applied:?}");
                    retry.remove(replicaset_id);
                }
            }
        }
    }

    fn release(&mut self) {
        log::info!(
            "release ddl lock for each instance within vshard group: {}",
            self.vshard_group_name
        );
        let topology = utils::tx_retry!(self.fetch_topology(), delay: 100, tries: 10, backoff: 2.0)
            .expect("failed to update topology on applier release");

        log::info!(
            "configure each instance with new topology after ddl applied within vshard group: {}",
            self.vshard_group_name
        );
        self.jobs_chan
            .send(types::Job::Configure(topology))
            .expect("sending to jobs chan should not fail");
        log::info!("wait for response from each instance that it saved new topology within vshard group: {}", self.vshard_group_name);
        self.configure_chan
            .recv()
            .expect("receiving from configure chan should not fail");
        self.ddl_cond.broadcast();
        log::info!(
            "released ddl lock for each instance from vshard group: {}",
            self.vshard_group_name
        );
    }

    fn fetch_topology(&mut self) -> Result<Vec<config::VShardGroupConfig>, error::Error> {
        log::info!(
            "fetch producer topology from url: {}",
            self.producer_cfg.topology_url
        );
        let response = self
            .client
            .get(&self.producer_cfg.topology_url)
            .expect("building get request should not fail")
            .send();
        let result = match response {
            Ok(mut v) => {
                let current_producer_topology = v
                    .json::<Vec<config::VShardGroupConfig>>()
                    .map_err(|err| {
                        log::error!("failed to decode producer topology: {err:?}");
                        error::Error::DecodeTopology
                    })?
                    .into_iter()
                    .map(|mut vshard_group| {
                        vshard_group.replicasets.retain(|r| r.is_storage);
                        vshard_group
                    })
                    .filter(|s| !s.replicasets.is_empty())
                    .collect();

                return Ok(current_producer_topology);
            }
            Err(e) => e,
        };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.fetch_topology();
        };
        log::error!("failed to fetch producer topology: {result:?}");
        Err(error::Error::FetchTopology)
    }

    pub fn fetch_space_info(
        &mut self,
        group: &String,
        space_name: &String,
    ) -> Result<Option<types::SpaceInfo>, error::Error> {
        let space_url = &self.producer_cfg.space_info_url;
        let response = self
            .client
            .get(format!("{space_url}/{group}/{space_name}"))
            .expect("building get request should not fail")
            .send();
        let result = match response {
            Ok(mut v) => {
                if v.status() != http_types::StatusCode::Ok {
                    return Err(error::Error::FetchSpaceInfo);
                }
                if let Ok(current_space_info) = v.json::<types::SpaceInfo>() {
                    return Ok(Some(current_space_info));
                }
                return Ok(None);
            }
            Err(e) => e,
        };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.fetch_space_info(group, space_name);
        };
        Err(error::Error::FetchSpaceInfo)
    }
}

pub struct Handle {
    close_chan: fiber::Channel<()>,
    ddl_chan: fiber::Channel<types::DDL>,
    ddl_cond: rc::Rc<fiber::Cond>,
    fiber: Option<fiber::JoinHandle<'static, ()>>,
}

impl Handle {
    pub fn apply(&self, ddl: types::DDL) -> rc::Rc<fiber::Cond> {
        self.ddl_chan
            .send(ddl)
            .expect("sending to ddl chan should not fail");
        self.ddl_cond.clone()
    }
}

impl Drop for Handle {
    fn drop(&mut self) {
        self.close_chan
            .send(())
            .expect("sending to close chan should not fail");
        self.fiber.take().unwrap().join();
    }
}
