use std::{collections, mem, rc, time};

use picoplugin::system::tarantool::{fiber, net_box};

use crate::{error, types};

pub struct Fiber {
    leader_id: String,

    vshard_groups: Vec<String>,
    disabled_spaces: Vec<String>,
    buckets_per_writer: u64,

    close_chan: fiber::Channel<()>,

    client: fibreq::Client,

    producer_instances: types::ProducerInstances,
    consumer_connections: types::ConsumerConnections,

    producer_cfg: config::TarantoolConfig,
    consumer_cfg: config::TarantoolConfig,

    previous_producer_topology: Vec<config::VShardGroupConfig>,
    previous_consumer_topology: Vec<config::VShardGroupConfig>,
    current_producer_topology: Vec<config::VShardGroupConfig>,
    current_consumer_topology: Vec<config::VShardGroupConfig>,

    stop_jobs: Vec<types::Job>,
    start_jobs: Vec<types::Job>,
    jobs_chan: fiber::Channel<types::Job>,

    current_cmp: collections::HashSet<String>,
    previous_cmp: collections::HashSet<String>,

    producer_masters_map: collections::HashMap<String, String>,
    consumer_masters_map: collections::HashMap<String, String>,
    current_consumer_map: collections::HashMap<String, config::VShardGroupConfig>,
    current_producer_map: collections::HashMap<String, config::VShardGroupConfig>,
}

impl Fiber {
    pub fn new(
        leader_id: String,
        vshard_groups: Vec<String>,
        disabled_spaces: Vec<String>,
        jobs_chan: fiber::Channel<types::Job>,
        producer_cfg: config::TarantoolConfig,
        consumer_cfg: config::TarantoolConfig,
        producer_instances: types::ProducerInstances,
        consumer_connections: types::ConsumerConnections,
        buckets_per_writer: u64,
    ) -> Self {
        let close_chan = fiber::Channel::new(1);

        let client = fibreq::ClientBuilder::new().build();

        let previous_producer_topology = Vec::new();
        let previous_consumer_topology = Vec::new();
        let current_producer_topology = Vec::new();
        let current_consumer_topology = Vec::new();

        let stop_jobs = Vec::new();
        let start_jobs = Vec::new();

        let current_cmp = collections::HashSet::new();
        let previous_cmp = collections::HashSet::new();

        let producer_masters_map = collections::HashMap::new();
        let consumer_masters_map = collections::HashMap::new();
        let current_consumer_map = collections::HashMap::new();
        let current_producer_map = collections::HashMap::new();

        Self {
            leader_id,
            vshard_groups,
            disabled_spaces,
            buckets_per_writer,
            close_chan,
            client,
            producer_instances,
            consumer_connections,
            producer_cfg,
            consumer_cfg,
            previous_producer_topology,
            previous_consumer_topology,
            current_producer_topology,
            current_consumer_topology,
            stop_jobs,
            start_jobs,
            jobs_chan,
            current_cmp,
            previous_cmp,
            producer_masters_map,
            consumer_masters_map,
            current_consumer_map,
            current_producer_map,
        }
    }
    pub fn start(mut self) -> Handle {
        let close_chan = self.close_chan.clone();
        log::warn!("start discovery fiber");
        let handle = fiber::Builder::new()
            .func(move || {
                utils::tx_retry!(self.update_producer_password(), delay: 100, tries: 10, backoff: 2.0).expect("failed to update producer password on discovery start");
                utils::tx_retry!(self.update_consumer_password(), delay: 100, tries: 10, backoff: 2.0).expect("failed to update consumer password on discovery start");
                loop {
                    if self.close_chan.try_recv().is_ok() {
                        return;
                    }
                    if self.fetch_producer_topology().is_err() {
                        continue;
                    };
                    if self.fetch_consumer_topology().is_err() {
                        continue;
                    };
                    let producer_topology_changed = self.producer_topology_changed();
                    let consumer_topology_changed = self.consumer_topology_changed();
                    if !producer_topology_changed && !consumer_topology_changed {
                        log::info!("topologies did not change, will not do anything");
                        fiber::sleep(time::Duration::from_millis(1000));
                        continue;
                    }
                    self.fill_maps();
                    self.disable_jobs();
                    if consumer_topology_changed {
                        self.handle_consumer_topology_changed();
                    }
                    if producer_topology_changed {
                        self.handle_producer_topology_changed();
                    }
                    self.run_jobs();
                    fiber::sleep(time::Duration::from_millis(1000));
                }
            })
            .name("uroboros_discovery")
            .start()
            .expect("starting discovery fiber should not fail");

        Handle {
            handle: Some(handle),
            close_chan,
        }
    }

    fn update_producer_password(&mut self) -> Result<(), error::Error> {
        log::warn!("update password on producer cluster");
        let response = self
            .client
            .post(&self.producer_cfg.user_url)
            .expect("building post request should not fail")
            .json(&types::UpdatePassword {
                new_pass: unsafe { config::PASSWORD.clone() },
            })
            .expect("building post request should not fail")
            .send();
        let Err(result) = response else { return Ok(()) };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.update_producer_password();
        };
        log::error!("failed to update producer password: {result:?}");
        Err(error::Error::UpdatePassword)
    }

    fn update_consumer_password(&mut self) -> Result<(), error::Error> {
        log::warn!("update password on consumer cluster");
        let response = self
            .client
            .post(&self.consumer_cfg.user_url)
            .expect("building post request should not fail")
            .json(&types::UpdatePassword {
                new_pass: unsafe { config::PASSWORD.clone() },
            })
            .expect("building post request should not fail")
            .send();
        let Err(result) = response else { return Ok(()) };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.update_producer_password();
        };
        log::error!("failed to update consumer password: {result:?}");
        Err(error::Error::UpdatePassword)
    }

    fn fetch_producer_topology(&mut self) -> Result<(), error::Error> {
        log::info!(
            "fetch producer topology from url: {}",
            self.producer_cfg.topology_url
        );
        let response = self
            .client
            .get(&self.producer_cfg.topology_url)
            .expect("building get request should not fail")
            .send();
        let result = match response {
            Ok(mut v) => {
                let current_producer_topology = v
                    .json::<Vec<config::VShardGroupConfig>>()
                    .map_err(|e| {
                        log::error!("failed to decode producer topology: {e:?}");
                        error::Error::DecodeTopology
                    })?
                    .into_iter()
                    .map(|mut vshard_group| {
                        vshard_group.replicasets.retain(|r| r.is_storage);
                        vshard_group
                    })
                    .filter(|s| !s.replicasets.is_empty())
                    .collect();
                let previous_producer_topology = mem::replace(
                    &mut self.current_producer_topology,
                    current_producer_topology,
                );
                self.previous_producer_topology = previous_producer_topology;
                return Ok(());
            }
            Err(e) => e,
        };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.fetch_producer_topology();
        };
        log::error!(
            "failed to fetch producer topology: {result:?}, url: {}",
            self.producer_cfg.topology_url
        );
        Err(error::Error::FetchTopology)
    }

    fn fetch_consumer_topology(&mut self) -> Result<(), error::Error> {
        log::info!(
            "fetch consumer topology from url: {}",
            self.consumer_cfg.topology_url
        );
        let response = self
            .client
            .get(&self.consumer_cfg.topology_url)
            .expect("building get request should not fail")
            .send();
        let result = match response {
            Ok(mut v) => {
                let current_consumer_topology = v
                    .json::<Vec<config::VShardGroupConfig>>()
                    .map_err(|e| {
                        log::error!("failed to decode consumer topology: {e:?}");
                        error::Error::DecodeTopology
                    })?
                    .into_iter()
                    .map(|mut vshard_group| {
                        vshard_group.replicasets.retain(|r| r.is_storage);
                        vshard_group
                    })
                    .filter(|s| !s.replicasets.is_empty())
                    .collect();
                let previous_consumer_topology = mem::replace(
                    &mut self.current_consumer_topology,
                    current_consumer_topology,
                );
                self.previous_consumer_topology = previous_consumer_topology;
                return Ok(());
            }
            Err(e) => e,
        };
        if matches!(result, fibreq::Error::SocketClosed) {
            return self.fetch_consumer_topology();
        };
        log::error!(
            "failed to fetch consumer topology: {result:?}, url: {}",
            self.consumer_cfg.topology_url
        );
        Err(error::Error::FetchTopology)
    }

    fn consumer_topology_changed(&mut self) -> bool {
        log::info!("compare topologies for consumer");
        for before_group in &self.previous_consumer_topology {
            for before_replicaset in &before_group.replicasets {
                for before_server in &before_replicaset.servers {
                    if before_server.is_active_master {
                        self.previous_cmp.insert(format!(
                            "{}_{}",
                            &before_replicaset.uuid, &before_server.uuid,
                        ));
                    }
                }
            }
        }
        for after_group in &self.current_consumer_topology {
            for after_replicaset in &after_group.replicasets {
                for after_server in &after_replicaset.servers {
                    if after_server.is_active_master {
                        self.current_cmp
                            .insert(format!("{}_{}", &after_replicaset.uuid, &after_server.uuid));
                    }
                }
            }
        }
        let result = self.previous_cmp.ne(&self.current_cmp);
        self.previous_cmp.clear();
        self.current_cmp.clear();
        result
    }
    fn producer_topology_changed(&mut self) -> bool {
        log::warn!("compare topologies for producer");
        for before_group in &self.previous_producer_topology {
            for before_replicaset in &before_group.replicasets {
                for before_server in &before_replicaset.servers {
                    if before_server.is_active_master {
                        self.previous_cmp.insert(format!(
                            "{}_{}",
                            &before_replicaset.uuid, &before_server.uuid,
                        ));
                    }
                }
            }
        }
        for after_group in &self.current_producer_topology {
            for after_replicaset in &after_group.replicasets {
                for after_server in &after_replicaset.servers {
                    if after_server.is_active_master {
                        self.current_cmp
                            .insert(format!("{}_{}", &after_replicaset.uuid, &after_server.uuid));
                    }
                }
            }
        }
        let result = self.previous_cmp.ne(&self.current_cmp);
        self.previous_cmp.clear();
        self.current_cmp.clear();
        result
    }

    fn fill_maps(&mut self) {
        self.current_consumer_map.clear();
        self.current_producer_map.clear();
        for consumer_vshard_group in &self.current_consumer_topology {
            self.current_consumer_map.insert(
                consumer_vshard_group.name.clone(),
                consumer_vshard_group.clone(),
            );
        }

        for producer_vshard_group in &self.current_producer_topology {
            self.current_producer_map.insert(
                producer_vshard_group.name.clone(),
                producer_vshard_group.clone(),
            );
        }
    }

    fn disable_jobs(&mut self) {
        'outer: for previous_producer_vshard_group in &self.previous_producer_topology {
            log::info!(
                "check group to disable jobs: {}",
                previous_producer_vshard_group.name
            );
            let vshard_group = self
                .current_producer_map
                .get(&previous_producer_vshard_group.name);
            // Stop all replicators in case of vshard group was removed
            if vshard_group.is_none() {
                for replicaset in &previous_producer_vshard_group.replicasets {
                    log::info!(
                        "(1) add stop job for group: {}, rs: {}",
                        previous_producer_vshard_group.name,
                        replicaset.uuid
                    );
                    self.stop_jobs.push(types::Job::Stop {
                        vshard_group: previous_producer_vshard_group.name.clone(),
                        replicaset_id: replicaset.uuid.clone(),
                    });
                }
                continue 'outer;
            }

            let vshard_group = vshard_group.unwrap();

            // Stop all replicators in case of vshard group buckets were changed
            if vshard_group.bucket_count != previous_producer_vshard_group.bucket_count {
                for replicaset in &previous_producer_vshard_group.replicasets {
                    log::info!(
                        "(2) add stop job for group: {}, rs: {}",
                        previous_producer_vshard_group.name,
                        replicaset.uuid
                    );
                    self.stop_jobs.push(types::Job::Stop {
                        vshard_group: previous_producer_vshard_group.name.clone(),
                        replicaset_id: replicaset.uuid.clone(),
                    });
                }
                continue 'outer;
            }

            // Stop distinct replicaset which was removed
            for previous_replicaset in &previous_producer_vshard_group.replicasets {
                let mut is_presented = false;
                'inner: for current_replicasset in &vshard_group.replicasets {
                    if current_replicasset.uuid == previous_replicaset.uuid {
                        is_presented = true;
                        break 'inner;
                    }
                }
                if !is_presented {
                    log::info!(
                        "(3) add stop job for group: {}, rs: {}",
                        previous_producer_vshard_group.name,
                        previous_replicaset.uuid
                    );
                    self.stop_jobs.push(types::Job::Stop {
                        vshard_group: previous_producer_vshard_group.name.clone(),
                        replicaset_id: previous_replicaset.uuid.clone(),
                    });
                }
            }
        }
    }

    fn handle_consumer_topology_changed(&mut self) {
        log::warn!("consumer topology has changed");

        let mut consumer_guard = self.consumer_connections.lock();
        let mut writers_cfgs = Vec::with_capacity(self.current_consumer_topology.len());

        'outer: for consumer_vshard_group in &self.current_consumer_topology {
            if !self.vshard_groups.contains(&consumer_vshard_group.name) {
                log::info!(
                    "consumer vshard group is not in whitelist: {}",
                    consumer_vshard_group.name
                );
                continue 'outer;
            }

            if !self
                .current_producer_map
                .contains_key(consumer_vshard_group.name.as_str())
            {
                log::info!(
                    "consumer vshard group is not in producer: {}",
                    consumer_vshard_group.name
                );
                continue 'outer;
            }

            writers_cfgs.push(config::WriterConfig::Tarantool(
                config::TarantoolWriterConfig {
                    vshard: consumer_vshard_group.clone(),
                    buckets_per_writer: self.buckets_per_writer,
                },
            ));

            if consumer_guard.get(&consumer_vshard_group.name).is_none() {
                consumer_guard.insert(
                    consumer_vshard_group.name.clone(),
                    collections::HashMap::with_capacity(32),
                );
            }
            let consumer_vshard_connections = consumer_guard
                .get_mut(&consumer_vshard_group.name)
                .expect("consumer vshard group connections should exist");

            'inner: for replicaset in &consumer_vshard_group.replicasets {
                log::info!(
                    "might recreate connection for vshard group: {}, replicaset: {}",
                    consumer_vshard_group.name,
                    replicaset.uuid
                );
                let replicaset_master = replicaset.master();
                let would_recreate_connection = consumer_vshard_connections
                    .get(&replicaset.uuid)
                    .map_or(Some(true), |conn :&rc::Rc<net_box::Conn>| {
                        self.consumer_masters_map
                            .get(&replicaset.uuid)
                            .map(|v| {
                                if v == &replicaset_master.uuid {
                                    utils::tx_retry!(conn.ping(&net_box::Options::default()), delay: 500, tries: 10, backoff: 2.0).is_err()
                                } else {
                                    true
                                }
                            })
                    })
                    .unwrap_or(true);
                if !would_recreate_connection {
                    log::info!(
                        "will not recreate connection for vshard group: {}, replicaset: {}",
                        consumer_vshard_group.name,
                        replicaset.uuid
                    );

                    continue 'inner;
                };
                let Ok(connection) = utils::tx_retry!(connect_replicaset(replicaset), delay: 100, tries: 3, backoff: 2.0)
                else {
                    log::error!(
                        "failed to recreate connection for vshard group: {}, replicaset: {}",
                        consumer_vshard_group.name,
                        replicaset.uuid
                    );
                    break 'outer;
                };
                consumer_vshard_connections
                    .insert(replicaset.uuid.clone(), rc::Rc::new(connection));
                self.consumer_masters_map
                    .insert(replicaset.uuid.clone(), replicaset_master.uuid.clone());
            }
        }
        self.start_jobs
            .push(types::Job::CreateWriters { writers_cfgs });
    }
    fn handle_producer_topology_changed(&mut self) {
        log::warn!("producer topology has changed");
        let mut producer_guard = self.producer_instances.lock();

        'outer: for producer_vshard_group in &self.current_producer_topology {
            // Replicate only groups from config
            if !self.vshard_groups.contains(&producer_vshard_group.name) {
                log::info!(
                    "vshard group is not in whitelist: {}",
                    producer_vshard_group.name
                );
                continue 'outer;
            }

            producer_guard.insert(
                producer_vshard_group.name.clone(),
                producer_vshard_group
                    .replicasets
                    .iter()
                    .filter(|x| x.is_storage)
                    .count(),
            );

            // Replicate only same vshard groups
            let Some(consumer_vshard_group) = self
                .current_consumer_map
                .get(producer_vshard_group.name.as_str())
            else {
                log::info!(
                    "producer vshard group is not in consumer: {}",
                    producer_vshard_group.name
                );
                continue 'outer;
            };
            // VShard groups but have same quantity of buckets in total
            if producer_vshard_group.bucket_count != consumer_vshard_group.bucket_count {
                log::info!(
                    "vshard group bad buckets: {}, producer num: {}, consumer num: {}",
                    producer_vshard_group.name,
                    producer_vshard_group.bucket_count,
                    consumer_vshard_group.bucket_count
                );
                continue 'outer;
            }
            'inner: for replicaset in &producer_vshard_group.replicasets {
                if !replicaset.is_storage {
                    log::info!(
                        "skip vshard group replicaset: {}, {}",
                        producer_vshard_group.name,
                        replicaset.uuid
                    );
                    continue 'inner;
                }
                log::info!(
                    "might create start job for vshard group: {}, replicaset: {}",
                    producer_vshard_group.name,
                    replicaset.uuid
                );

                let replicaset_master = replicaset.master();
                let replicaset_master_changed = self
                    .producer_masters_map
                    .get(&replicaset.uuid)
                    .map_or(true, |v| v != &replicaset_master.uuid);
                if !replicaset_master_changed {
                    log::info!(
                        "might create start job for vshard group: {}, replicaset: {} cause master is still same",
                        consumer_vshard_group.name,
                        replicaset.uuid
                    );
                    continue 'inner;
                };
                self.producer_masters_map
                    .insert(replicaset.uuid.clone(), replicaset_master.uuid.clone());
                log::info!(
                    "create start job for vshard group: {}, replicaset: {}",
                    producer_vshard_group.name,
                    replicaset.uuid
                );
                self.start_jobs.push(types::Job::Start {
                    config: Box::new(config::JobConfig {
                        leader_id: self.leader_id.clone(),
                        replicaset_id: replicaset.uuid.clone(),
                        replicaset_alias: replicaset.alias.clone(),
                        vshard_group: producer_vshard_group.name.clone(),
                        disabled_spaces: self.disabled_spaces.clone(),
                        reader: config::ReaderConfig::from(replicaset.clone()),
                    }),
                    topology: self.current_producer_topology.clone(),
                });
            }
        }
    }

    fn run_jobs(&mut self) {
        log::info!("sending stop jobs: {:?}", self.stop_jobs);
        for stop_job in self.stop_jobs.drain(0..) {
            self.jobs_chan
                .send(stop_job)
                .expect("sending to jobs channel should not fail");
        }
        log::info!("sending start jobs: {:?}", self.start_jobs);
        for start_job in self.start_jobs.drain(0..) {
            self.jobs_chan
                .send(start_job)
                .expect("sending to jobs channel should not fail");
        }
    }
}

pub struct Handle {
    close_chan: fiber::Channel<()>,
    handle: Option<fiber::JoinHandle<'static, ()>>,
}

impl Drop for Handle {
    fn drop(&mut self) {
        self.close_chan
            .send(())
            .expect("sending to close channel should not fail");
        self.handle.take().unwrap().join();
    }
}

fn connect_replicaset(shard: &config::ReplicasetConfig) -> Result<net_box::Conn, error::Error> {
    let master = shard.master();
    let (username, password, host, port) = master.parse_uri();
    let addr = format!("{host}:{port}");
    log::info!("connecting to replicaset: {}, addr: {addr}", shard.uuid);
    let instance = match net_box::Conn::new(
        &addr,
        net_box::ConnOptions {
            user: username.to_string(),
            password: password.to_string(),
            reconnect_after: time::Duration::from_millis(100),
            ..net_box::ConnOptions::default()
        },
        None,
    ) {
        Ok(v) => v,
        Err(e) => {
            log::error!("failed to connect replicaset: {} with: {:?}", addr, e);
            return Err(error::Error::ConnectReplicaset);
        }
    };
    if let Err(e) = instance.wait_connected(None) {
        log::error!("failed to wait connect replicaset: {} with: {:?}", addr, e);
        return Err(error::Error::ConnectReplicaset);
    }
    Ok(instance)
}
