#![allow(clippy::new_without_default)]
#![allow(clippy::upper_case_acronyms)]
#![allow(clippy::too_many_arguments)]
#![allow(clippy::too_many_lines)]

use picoplugin::system::tarantool::fiber;
use std::rc;

mod error;
mod fibers;
mod observer;
mod types;
mod uroboros;
pub use types::{SidecarDDL, SidecarDDLType, DDL};

/// Coordinator of uroboros cluster.
pub struct Coordinator<'a> {
    id: String,
    uroboros: Option<uroboros::Uroboros<'a>>,
    observer: Option<Box<dyn observer::Observer>>,
    config: Option<config::Config>,
}

impl<'a> Coordinator<'a> {
    /// Creates a new empty instance of `Coordinator`.
    ///
    /// # Returns
    ///
    /// Not configured `Coordinator`.
    #[must_use]
    pub fn new() -> Self {
        Self {
            uroboros: None,
            observer: None,
            config: None,
            id: String::new(),
        }
    }

    pub fn configure(&mut self, config: config::Config, id: String) {
        log::warn!("configure coordinator");
        self.id = id;
        self.config = Some(config);
    }

    /// Starts an uroboros coordinator, creating [`observer::Observer`] for consumer and producer cluster
    /// along with [`uroboros::Uroboros`] job manager.
    ///
    /// # Panics
    ///
    /// - In case of [`config::Config`] is not set.
    pub fn start(&mut self) {
        log::warn!("start coordinator");
        let config = self
            .config
            .clone()
            .expect("get config should not fail on coordinator start");

        let jobs_ch = fiber::Channel::new(1);
        let configure_chan = fiber::Channel::new(1);

        let observer = match config.consumer {
            config::ConsumerConfig::Kafka(_) => unimplemented!(),
            config::ConsumerConfig::Tarantool(tnt) => Box::new(observer::Tarantool::new(
                &config.enabled_groups,
                config.disabled_spaces.clone(),
                self.id.clone(),
                &jobs_ch,
                &config.producer,
                &tnt,
                config.buckets_per_writer,
                &configure_chan,
            )) as Box<dyn observer::Observer>,
        };

        let own = uroboros::Uroboros::new(jobs_ch, configure_chan);

        self.observer = Some(observer);
        self.uroboros = Some(own);
    }

    pub fn stop(&mut self) {
        log::warn!("stop coordinator");
        if let Some(own) = self.uroboros.take() {
            drop(own);
        }
        if let Some(observer) = self.observer.take() {
            drop(observer);
        }
    }

    /// Handles received ddl. It will block requesting instance of uroboros worker unless
    /// each of them sends it's ddl.
    ///
    /// After that ddl will be applied to target cluster, and each instance of uroboros worker will
    /// be allowed to continue replication process.
    ///
    /// # Panics
    ///
    /// - In case of there are somehow no observer. But it must have been setted up on start.
    ///
    /// # Returns
    ///
    /// A fiber cond which will hold the lock untill each replicator sends it's ddl.
    pub fn handle_ddl(&mut self, ddl: types::DDL) -> rc::Rc<fiber::Cond> {
        log::warn!("ddl coordinator");
        self.observer
            .as_ref()
            .expect("getting observer on handle ddl should not fail")
            .apply_ddl(ddl)
    }
}
