use crate::{fibers, types};
use picoplugin::system::tarantool::fiber;
use std::{collections, rc};

pub trait Observer {
    fn apply_ddl(&self, ddl: types::DDL) -> rc::Rc<fiber::Cond>;
}

pub struct Tarantool {
    _discovery_handle: fibers::DiscoveryHandle,
    applier_handles: collections::HashMap<String, fibers::ApplierHandle>,
}

impl Observer for Tarantool {
    fn apply_ddl(&self, ddl: types::DDL) -> rc::Rc<fiber::Cond> {
        log::info!("apply ddl for group: {}", ddl.vshard_group);
        let handle = self
            .applier_handles
            .get(&ddl.vshard_group)
            .expect("vshard group applier handle should exist");
        handle.apply(ddl)
    }
}

impl Tarantool {
    pub fn new(
        vshard_groups: &[String],
        disabled_spaces: Vec<String>,
        leader_id: String,
        jobs_chan: &fiber::Channel<types::Job>,
        producer_cfg: &config::TarantoolConfig,
        consumer_cfg: &config::TarantoolConfig,
        buckets_per_writer: u64,
        configure_chan: &fiber::Channel<()>,
    ) -> Self {
        let consumer_connections =
            rc::Rc::new(fiber::Mutex::new(collections::HashMap::with_capacity(32)));
        let producer_instances =
            rc::Rc::new(fiber::Mutex::new(collections::HashMap::with_capacity(32)));

        let discovery_handle = fibers::DiscoveryFiber::new(
            leader_id,
            vshard_groups.to_vec(),
            disabled_spaces,
            jobs_chan.clone(),
            producer_cfg.clone(),
            consumer_cfg.clone(),
            producer_instances.clone(),
            consumer_connections.clone(),
            buckets_per_writer,
        )
        .start();

        let mut applier_handles = collections::HashMap::with_capacity(vshard_groups.len());

        for vshard_group_name in vshard_groups {
            let applier_handle = fibers::ApplierFiber::new(
                vshard_group_name.clone(),
                producer_instances.clone(),
                consumer_connections.clone(),
                jobs_chan.clone(),
                configure_chan.clone(),
                producer_cfg.clone(),
            )
            .start();
            applier_handles.insert(vshard_group_name.clone(), applier_handle);
        }

        Self {
            applier_handles,
            _discovery_handle: discovery_handle,
        }
    }
}
