use std::{collections, time};

use picoplugin::system::tarantool::fiber;
use picoplugin::transport::rpc;

use crate::types::{self};

const GET_ISNTANCES_QUERY: &str =
    "SELECT \"instance_id\", \"instance_uuid\" FROM \"_pico_instance\";";

pub struct Uroboros<'a> {
    close_ch: fiber::Channel<()>,
    runner: Option<fiber::JoinHandle<'a, ()>>,
}

impl<'a> Uroboros<'a> {
    /// Creates a new instance of `Uroboros` cluster watcher.
    ///
    /// # Parameters
    ///
    /// - `jobs_ch`: Channel which is used to apply jobs to `Uroboros` cluster.
    ///
    /// # Returns
    ///
    /// A new instance of `Uroboros`.
    pub fn new(jobs_ch: fiber::Channel<types::Job>, configure_ch: fiber::Channel<()>) -> Self {
        let close_ch = fiber::Channel::new(1);

        let rows = utils::tx_retry!(
            picoplugin::sql::query(GET_ISNTANCES_QUERY).fetch::<types::Instance>(),
            delay: 100, tries: 10, backoff: 2.0
        )
        .unwrap();

        let mut instances = Vec::with_capacity(rows.len());

        for instance in rows {
            instances.push(instance.instance_id);
        }

        let runner = Some({
            let close_ch = close_ch.clone();

            fiber::Builder::new()
                .func(move || {
                    let mut index = 0;
                    let pkg_version = env!("CARGO_PKG_VERSION");
                    let mut map = collections::HashMap::new();

                    loop {
                        if close_ch.try_recv().is_ok() {
                            return;
                        }
                        if jobs_ch.is_empty() {
                            fiber::sleep(time::Duration::from_millis(100));
                            continue;
                        }

                        let mut next_instance_id = || {
                            let instance_id = if let Some(v) = instances.get(index) {
                                v
                            } else {
                                index = 0;
                                instances.get(index).unwrap()
                            };

                            index += 1;
                            instance_id
                        };

                        match jobs_ch.recv().expect("recv from jobs chan should not fail") {
                            types::Job::Start { config, topology } => {
                                log::info!(
                                    "received start job: {config:?}, topology: {topology:?}"
                                );
                                let instance_id = next_instance_id();

                                map.insert(config.replicaset_id.clone(), instance_id.clone());
                                log::warn!(
                                    "call configure on start job for instance_id: {instance_id}"
                                );
                                utils::tx_retry!(rpc::RequestBuilder::new(
                                    rpc::RequestTarget::InstanceId(instance_id)
                                )
                                .path(config::RPC_REPLICATOR_CONFIGURE_ROUTE)
                                .plugin_service("uroboros", "uroboros")
                                .plugin_version(pkg_version)
                                .input(
                                    rpc::Request::encode_rmp(&topology)
                                        .expect("encoding topology with rmp should not fail")
                                )
                                .send(), delay: 100, tries: 10, backoff: 2.0)
                                .expect("sending replicator configure should not fail");
                                log::warn!("call add on start job for instance_id: {instance_id}");
                                utils::tx_retry!(rpc::RequestBuilder::new(
                                    rpc::RequestTarget::InstanceId(instance_id)
                                )
                                .path(config::RPC_REPLICATOR_ADD_ROUTE)
                                .plugin_service("uroboros", "uroboros")
                                .plugin_version(pkg_version)
                                .input(
                                    rpc::Request::encode_rmp(&config)
                                        .expect("encoding config with rmp should not fail")
                                )
                                .send(), delay: 100, tries: 10, backoff: 2.0)
                                .expect("sending replicator add should not fail");
                                log::warn!(
                                    "sent job to start worker for group {}, replicaset {}",
                                    config.vshard_group,
                                    config.replicaset_id,
                                );
                            }
                            types::Job::Stop {
                                vshard_group,
                                replicaset_id,
                            } => {
                                log::warn!("received stop job for vshard_group: {vshard_group}, replicaset: {replicaset_id}");
                                let instance_id = map.remove(&replicaset_id).expect(
                                    "getting instance id for replicaset id should not fail",
                                );
                                log::warn!("call remove on stop job for instance_id: {instance_id}");
                                utils::tx_retry!(rpc::RequestBuilder::new(
                                    rpc::RequestTarget::InstanceId(&instance_id)
                                )
                                .path(config::RPC_REPLICATOR_REMOVE_ROUTE)
                                .plugin_service("uroboros", "uroboros")
                                .plugin_version(pkg_version)
                                .input(
                                    rpc::Request::encode_rmp(&(&vshard_group, &replicaset_id))
                                        .expect("encoding config with rmp should not fail")
                                )
                                .send(), delay: 100, tries: 10, backoff: 2.0)
                                .expect("sending replicator remove should not fail");
                            }
                            types::Job::Configure(topology) => {
                                log::info!("received configure job with topology: {topology:?}");
                                for instance_id in &instances {
                                    log::warn!("call configure on configure job for instance_id: {instance_id}");
                                    utils::tx_retry!(rpc::RequestBuilder::new(
                                        rpc::RequestTarget::InstanceId(instance_id)
                                    )
                                    .path(config::RPC_REPLICATOR_CONFIGURE_ROUTE)
                                    .plugin_service("uroboros", "uroboros")
                                    .plugin_version(pkg_version)
                                    .input(
                                        rpc::Request::encode_rmp(&topology)
                                            .expect("encoding topology with rmp should not fail")
                                    )
                                    .send(), delay: 100, tries: 10, backoff: 2.0)
                                    .expect("sending replicator configure should not fail");
                                }
                                log::warn!("send callback on configure");
                                configure_ch
                                    .send(())
                                    .expect("sending to configure channel should not fail");
                            }
                            types::Job::CreateWriters {
                                writers_cfgs,
                            } => {
                                for instance_id in &instances {
                                    log::warn!("call create writers for instance_id: {instance_id}");
                                    utils::tx_retry!(rpc::RequestBuilder::new(
                                        rpc::RequestTarget::InstanceId(instance_id)
                                    )
                                    .path(config::RPC_REPLICATOR_WRITERS_ROUTE)
                                    .plugin_service("uroboros", "uroboros")
                                    .plugin_version(pkg_version)
                                    .input(
                                        rpc::Request::encode_rmp(&writers_cfgs)
                                            .expect("encoding request with rmp should not fail")
                                    )
                                    .send(), delay: 100, tries: 10, backoff: 2.0)
                                    .expect("sending replicator create writers should not fail");
                                }
                            }
                        };
                    }
                })
                .name(String::from("uroboros_runner"))
                .start()
                .expect("starting runner fiber should not fail")
        });

        Self { close_ch, runner }
    }
}

impl<'a> Drop for Uroboros<'a> {
    fn drop(&mut self) {
        _ = self.close_ch.send(());
        if let Some(runner) = self.runner.take() {
            runner.join();
        }
    }
}
