use std::{
    collections::HashSet,
    mem, rc,
    time::{self, Instant},
};

use picoplugin::system::tarantool::{
    fiber,
    index::{IndexType, IteratorType},
    space::{FieldType, IsNullable, Space, SpaceType},
    tuple,
};

use crate::{data::state, error, metrics, types};

const QUEUE_SPACE_NAME: &str = "uroboros_lsn_queue";
const QUEUE_UNWRITTEN_INDEX: &str = "index_lsn_unwrited";

#[derive(Clone, Debug, serde::Serialize, serde::Deserialize)]
struct QueueItem {
    replicaset_uuid: String,
    is_written: bool,
    lsn: u64,
    replica_uuid: String,
    replica_id: u64,
    shard_group: String,
    replicaset_alias: Option<String>,
}

impl tuple::Encode for QueueItem {}

#[derive(Debug)]
pub enum LSNQueueMsg {
    Store {
        replica_info: rc::Rc<types::ReplicaInfo>,
        replica_id: u64,
        lsn: u64,
    },
    Written {
        replica_info: rc::Rc<types::ReplicaInfo>,
        replica_id: u64,
        lsn: u64,
    },
}

pub struct LSNQueue<'a> {
    lsn_chan: fiber::Channel<LSNQueueMsg>,
    _handle: Handle<'a>,
}

impl<'a> LSNQueue<'a> {
    /// Creates a temporary tarantool space for LSN queue
    /// and an `LSNQueue` item.
    ///
    /// # Panics
    ///
    /// - In case of failing creating space or its indexes.
    pub fn new(lsn_chan: fiber::Channel<LSNQueueMsg>) -> Self {
        log::info!("create space for lsn queue if it is not exist");
        let space = utils::tx_retry!(Space::builder(QUEUE_SPACE_NAME)
            .if_not_exists(true)
            .field(("replicaset_uuid", FieldType::String, IsNullable::NonNullalbe))
            .field(("is_written", FieldType::Boolean, IsNullable::NonNullalbe))
            .field(("lsn", FieldType::Integer, IsNullable::NonNullalbe))
            .field(("replica_uuid", FieldType::String, IsNullable::NonNullalbe))
            .field(("replica_id", FieldType::Unsigned, IsNullable::NonNullalbe))
            .field(("shard_group", FieldType::String, IsNullable::NonNullalbe))
            .field(("replicaset_alias", FieldType::String, IsNullable::Nullable))
            .space_type(SpaceType::Temporary)
            .create(), delay: 100, tries: 5, backoff: 2.0)
        .expect("creating lsn queue space should not fail");
        utils::tx_retry!(space
            .index_builder("primary")
            .if_not_exists(true)
            .index_type(IndexType::Tree)
            .parts(["replicaset_uuid", "lsn"])
            .create(), delay: 100, tries: 5, backoff: 2.0)
        .expect("creating index of lsn queue space should not fail");
        utils::tx_retry!(space
            .index_builder(QUEUE_UNWRITTEN_INDEX)
            .if_not_exists(true)
            .index_type(IndexType::Tree)
            .parts(["replicaset_uuid", "is_written", "lsn"])
            .create(), delay: 100, tries: 5, backoff: 2.0)
        .expect("creating index of lsn queue space should not fail");

        log::info!("start LSN queue");
        let close = fiber::Channel::new(1);
        let handle = {
            let close = close.clone();
            let fiber_lsn_chan = lsn_chan.clone();
            fiber::Builder::new()
                .func(move || {
                    let mut replicaset_uuids: HashSet<String> = HashSet::new();
                    let mut last_commit: Option<Instant> = None;
                    loop {
                        if close.try_recv().is_ok() {
                            log::info!("close LSN queue");
                            return;
                        }

                        if !fiber_lsn_chan.is_empty() {
                            if let Some(msg) = fiber_lsn_chan.recv() {
                                match msg {
                                    LSNQueueMsg::Store { replica_info, replica_id, lsn } => {
                                        while let Err(e) = store(
                                            replica_info.clone(), replica_id, lsn, false,
                                        ) {
                                            log::error!(
                                                "can not put lsn to queue (vshard group {}, replica_uuid {}): {e:?}",
                                                replica_info.vshard_group,
                                                replica_info.replica_uuid,
                                            );
                                            fiber::sleep(time::Duration::from_millis(100));
                                        }
                                    },
                                    LSNQueueMsg::Written { replica_info, replica_id, lsn } => {
                                        while let Err(e) = store(
                                            replica_info.clone(), replica_id, lsn, true,
                                        ) {
                                            log::error!(
                                                "can not update lsn in queue (vshard group {}, replica_uuid {}): {e:?}",
                                                replica_info.vshard_group,
                                                replica_info.replica_uuid,
                                            );
                                            fiber::sleep(time::Duration::from_millis(100));
                                        }
                                        if !replicaset_uuids.contains(&replica_info.replicaset_uuid) {
                                            replicaset_uuids.insert(replica_info.replicaset_uuid.clone());
                                        }
                                    },
                                }
                            }
                        }

                        let since_commit = if let Some(v) = last_commit {
                            v.elapsed()
                        } else {
                            Instant::now().checked_sub(time::Duration::from_secs(3)).expect("time substraction should not fail").elapsed()
                        };
                        if since_commit.as_secs() >= 2 {
                            for replicaset_uuid in &replicaset_uuids {
                                while let Err(e) = commit_lsn(replicaset_uuid) {
                                    log::error!(
                                        "can not update lsns for replica_id {replicaset_uuid}: {e:?}",
                                    );
                                    fiber::sleep(time::Duration::from_millis(100));
                                }
                            }
                            last_commit = Some(Instant::now());
                        }

                        if fiber_lsn_chan.is_empty() {
                            fiber::sleep(time::Duration::from_millis(100));
                        }
                    }
            })
                .name("lsn_queue")
                .start()
                .expect("spawning lsn queue fiber should not fail")
        };
        LSNQueue {
            lsn_chan,
            _handle: Handle {
                close,
                handle: mem::ManuallyDrop::new(handle),
            },
        }
    }

    pub fn channel(&self) -> fiber::Channel<LSNQueueMsg> {
        self.lsn_chan.clone()
    }
}

fn store(
    replica_info: rc::Rc<types::ReplicaInfo>,
    replica_id: u64,
    lsn: u64,
    is_written: bool,
) -> Result<(), error::Error> {
    log::info!(
        "store lsn to queue: {} - {lsn}",
        replica_info.replicaset_uuid,
    );
    let _ = utils::tx_retry!(Space::find(QUEUE_SPACE_NAME)
        .expect("space should exist on store")
        .replace(&(&replica_info.replicaset_uuid, is_written, lsn, &replica_info.replica_uuid, replica_id, &replica_info.vshard_group, &replica_info.replicaset_alias))
        .map_err(|e| {
            log::error!("failed to store lsn to queue: {e:?}");
            error::Error::LsnQueueCreateRecord
        }), delay: 100, tries: 10, backoff: 2.0)?;
    Ok(())
}

fn commit_lsn(replicaset_uuid: &str) -> Result<(), error::Error> {
    let queue = Space::find(QUEUE_SPACE_NAME).expect("space should exist on set flag");
    let mut to_commit = None;
    let mut to_delete = vec![];
    log::info!("get written LSNs for {replicaset_uuid}");
    let written = utils::tx_retry!(queue
        .index(QUEUE_UNWRITTEN_INDEX)
        .expect("index should exist")
        .select(IteratorType::GE, &(replicaset_uuid, &true))
        .map_err(|e| {
            log::error!("failed to get lsn queue for {replicaset_uuid}: {e:?}");
            error::Error::LsnQueueSetWritten
        }), delay: 100, tries: 10, backoff: 2.0)?;
    for tuple in written {
        let item: QueueItem = tuple.decode().map_err(|e| {
            log::error!("failed to decode record: {e:?}");
            error::Error::LsnQueueSetWritten
        })?;
        log::info!("decode lsn item operation: {item:?}");
        if item.replicaset_uuid != replicaset_uuid || !item.is_written {
            break;
        }
        to_commit = Some(item.clone());
        to_delete.push(item);
    }
    if let Some(item) = to_commit {
        let replicaset_uuid = item.replicaset_uuid.clone();
        utils::tx_retry!(state::update(&state::StoredItem {
            cluster_uuid: replicaset_uuid.to_string(),
            replica_uuid: item.replica_uuid.to_string(),
            replica_id: item.replica_id,
            lsn: item.lsn,
        }), delay: 100, tries: 10, backoff: 2.0)?;
        metrics::LSN_GAUGE
            .with_label_values(&[
                &item.shard_group,
                &item.replicaset_uuid,
                item.replicaset_alias.clone().unwrap_or_default().as_str(),
            ])
            .set(item.lsn as f64);
        for op in to_delete {
            utils::tx_retry!(queue.delete(&(op.replicaset_uuid.clone(), op.lsn)).map_err(|e| {
                log::error!("failed to delete lsn from queue: {e:?}");
                error::Error::LsnQueueSetWritten
            }), delay: 100, tries: 10, backoff: 2.0)?;
        }
    }
    Ok(())
}

struct Handle<'a> {
    handle: mem::ManuallyDrop<fiber::JoinHandle<'a, ()>>,
    close: fiber::Channel<()>,
}

impl<'a> Drop for Handle<'a> {
    fn drop(&mut self) {
        self.close
            .send(())
            .expect("sending to close channel should not fail");
        unsafe {
            mem::ManuallyDrop::take(&mut self.handle).join();
        }
    }
}
