use picoplugin::sql::{self, types};

use crate::error;

const GET_STATE_QUERY: &str = r#"SELECT "cluster_uuid", "replica_uuid", "replica_id", "lsn" FROM "state" WHERE "cluster_uuid" = ?;"#;
const INSERT_STATE_QUERY: &str = r#"INSERT INTO "state" ("cluster_uuid", "replica_uuid", "replica_id", "lsn") VALUES (?, ?, ?, ?);"#;
const UPDATE_STATE_QUERY: &str = r#"UPDATE "state" SET "lsn"  = ? WHERE  "cluster_uuid" = ? AND "replica_uuid" = ? AND "replica_id" = ?;"#;

#[derive(Clone, Debug, Default, serde::Serialize, serde::Deserialize)]

pub struct StoredItem {
    pub cluster_uuid: String,
    pub replica_uuid: String,
    pub replica_id: u64,
    pub lsn: u64,
}

pub fn select(cluster_uuid: &str) -> Result<Option<Vec<StoredItem>>, error::Error> {
    log::info!("select state for cluster: {cluster_uuid}");
    let rows = sql::query(GET_STATE_QUERY)
        .bind(types::SqlValue::string(cluster_uuid))
        .fetch::<StoredItem>()
        .map_err(|e| {
            log::error!("failed to select state: {e:?}");
            error::Error::SqlExecute
        })?;
    if rows.is_empty() {
        Ok(None)
    } else {
        Ok(Some(rows))
    }
}

pub fn create(state: &crate::types::State) -> Result<(), error::Error> {
    log::info!("state to insert: {:?}", state);
    for (i, v) in state.vclock.iter() {
        sql::query(INSERT_STATE_QUERY)
            .bind(types::SqlValue::string(state.cluster_uuid.to_string()))
            .bind(types::SqlValue::string(state.replica_uuid.to_string()))
            .bind(types::SqlValue::unsigned(*i))
            .bind(types::SqlValue::unsigned(*v))
            .execute()
            .map_err(|e| {
                log::error!("failed to create state: {e:?}");
                error::Error::SqlExecute
            })?;
    }
    Ok(())
}

pub fn update(item: &StoredItem) -> Result<(), error::Error> {
    log::info!("update state for cluster: {}", item.cluster_uuid);
    sql::query(UPDATE_STATE_QUERY)
        .bind(types::SqlValue::unsigned(item.lsn))
        .bind(types::SqlValue::string(&item.cluster_uuid))
        .bind(types::SqlValue::string(&item.replica_uuid))
        .bind(types::SqlValue::unsigned(item.replica_id))
        .execute()
        .map_err(|e| {
            log::error!("failed to update state: {e:?}");
            error::Error::SqlExecute
        })?;
    Ok(())
}
