use std::rc;

use crate::{config, error, package, sidecar};
use picoplugin::system::tarantool::{lua_state, tlua};

const VSHARD_INIT_FUNCTION: &str = "
    function init_vshard(router_name, cfg_string) 
        local full_router_name = 'uroboros_' .. router_name
        local router = _G[full_router_name]
        local cfg, number = require('msgpack').decode(cfg_string)
        if router == nil then
            _G[full_router_name] = require('vshard').router.new(full_router_name, cfg)
        else
            router:cfg(cfg)
        end
        return 0
    end
";

const APPLY_DML_DELETE_FUNCTION: &str = "
function apply_dml_delete(router_name, bucket_id, ...)
    local full_router_name = 'uroboros_' .. router_name
    local _, error =  _G[full_router_name]:call(bucket_id,
        { mode = 'write' },
        'liburoboros_apply_dml_delete',
        {...},
        { timeout = 5 }
    )

    return 0
end
";

const APPLY_DML_REPLACE_FUNCTION: &str = "
function apply_dml_replace(router_name, bucket_id, ...)
    local full_router_name = 'uroboros_' .. router_name
    local _, error =  _G[full_router_name]:call(bucket_id,
        { mode = 'write' },
        'liburoboros_apply_dml_replace',
        {...},
        { timeout = 5 }
    )

    return 0
end
";

const INIT_VSHARD_NAME: &str = "init_vshard";
const APPLY_DML_DELETE_NAME: &str = "apply_dml_delete";
const APPLY_DML_REPLACE_NAME: &str = "apply_dml_replace";

pub fn init() {
    let lua = lua_state();

    lua.exec(VSHARD_INIT_FUNCTION)
        .expect("creating vshard init function should not fail");
    lua.exec(APPLY_DML_DELETE_FUNCTION)
        .expect("creating apply dml function should not fail");
    lua.exec(APPLY_DML_REPLACE_FUNCTION)
        .expect("creating apply dml function should not fail");
}

pub fn init_vshard(cfg: config::VShardGroupConfig) -> Result<(), error::Error> {
    let lua = lua_state();

    log::info!("init vshard with cfg: {cfg:?}");
    let apply_dml_f = lua
        .get::<tlua::LuaFunction<_>, _>(INIT_VSHARD_NAME)
        .expect("getting init vshard function should not fail");
    let name = cfg.name.clone();
    apply_dml_f
        .call_with_args::<u64, _>(&(&name, cfg.as_table()))
        .map_err(|e| {
            log::error!("failed init vshard: {e:?}");
            error::Error::LuaInitVshard
        })?;
    Ok(())
}

pub fn apply_dml(dml: rc::Rc<package::DML>) -> Result<(), error::Error> {
    let lua = lua_state();

    let group = dml.meta.vshard_group.clone();
    let bucket_id = dml.meta.bucket_id.unwrap();
    let dml: sidecar::DML = (&*dml).into();

    match dml {
        sidecar::DML::Delete(v) => {
            let apply_dml_f = lua
                .get::<tlua::LuaFunction<_>, _>(APPLY_DML_DELETE_NAME)
                .expect("getting apply dml function should not fail");
            apply_dml_f
                .call_with_args::<u64, _>(&(&group, bucket_id, v))
                .map_err(|e| {
                    log::error!(
                        "failed apply dml for bucket_id {bucket_id} in group {group} with: {e:?}"
                    );
                    error::Error::LuaApplyDML
                })?;
        }
        sidecar::DML::Replace(v) => {
            let apply_dml_f = lua
                .get::<tlua::LuaFunction<_>, _>(APPLY_DML_REPLACE_NAME)
                .expect("getting apply dml function should not fail");
            apply_dml_f
                .call_with_args::<u64, _>(&(&group, bucket_id, v))
                .map_err(|e| {
                    log::error!(
                        "failed apply dml for bucket_id {bucket_id} in group {group} with: {e:?}"
                    );
                    error::Error::LuaApplyDML
                })?;
        }
    }

    Ok(())
}
