use crate::config as conf;
use picomole::package;

/// Helper struct to make correct convertion into DDL using
/// provided vshard group and replicaset id.
#[derive(Debug)]
pub struct IRDDL<'a> {
    dml: package::DML,
    replicaset_id: &'a str,
    vshard_group: &'a str,
}

impl<'a> IRDDL<'a> {
    pub fn new(dml: package::DML, replicaset_id: &'a str, vshard_group: &'a str) -> Self {
        Self {
            dml,
            replicaset_id,
            vshard_group,
        }
    }
}

#[derive(Debug)]
pub struct Meta {
    pub space_id: u32,
    pub lsn: Option<u64>,
    pub replica_id: Option<u64>,
    pub replicaset_id: String,
    pub vshard_group: String,
    pub target_space_id: u32,
}

#[derive(Debug)]
pub enum Operation {
    Insert { tuple: rmpv::Value },
    Delete { key: rmpv::Value },
    Update { key: rmpv::Value, ops: rmpv::Value },
}

#[derive(Debug)]
pub struct DDL {
    pub meta: Meta,
    pub operation: Operation,
}

impl<'a> TryFrom<IRDDL<'a>> for DDL {
    type Error = package::DML;
    fn try_from(value: IRDDL<'a>) -> Result<Self, Self::Error> {
        log::info!("try to convert DML to DDL: {value:?}");
        let ddl = match value.dml {
            package::DML::Insert(v) => {
                let space_id =
                    u32::try_from(v.body().space_id).expect("casting u64 to u32 should not fail");
                if !super::is_system_space(space_id) {
                    log::info!(
                        "cannot cast DML to DDL cause its not system space with id: {space_id}"
                    );
                    return Err(package::DML::Insert(v));
                }
                if super::is_ignored_system_space(space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its ignored system space with id: {space_id}"
                    );
                    return Err(package::DML::Insert(v));
                }
                let tuple = v.body().tuple.clone();
                let target_space_id =
                    u32::try_from(tuple.as_array().unwrap().first().unwrap().as_u64().unwrap())
                        .expect("casting u64 to u32 should not fail");
                if super::is_system_space(target_space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its DDL onto system space with id: {target_space_id}"
                    );
                    return Err(package::DML::Insert(v));
                }
                let meta = Meta {
                    space_id,
                    lsn: v.header().lsn,
                    target_space_id,
                    replica_id: v.header().replica_id,
                    replicaset_id: value.replicaset_id.to_string(),
                    vshard_group: value.vshard_group.to_string(),
                };
                let operation = Operation::Insert { tuple };
                DDL { meta, operation }
            }
            package::DML::Update(v) => {
                let space_id =
                    u32::try_from(v.body().space_id).expect("casting u64 to u32 should not fail");
                if !super::is_system_space(space_id) {
                    log::info!(
                        "cannot cast DML to DDL cause its not system space with id: {space_id}"
                    );
                    return Err(package::DML::Update(v));
                }
                if super::is_ignored_system_space(space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its ignored system space with id: {space_id}"
                    );
                    return Err(package::DML::Update(v));
                }
                let target_space_id = u32::try_from(
                    v.body()
                        .key
                        .as_array()
                        .unwrap()
                        .first()
                        .unwrap()
                        .as_u64()
                        .unwrap(),
                )
                .expect("casting u64 to u32 should not fail");

                if super::is_system_space(target_space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its DDL onto system space with id: {target_space_id}"
                    );
                    return Err(package::DML::Update(v));
                }

                let meta = Meta {
                    space_id,
                    lsn: v.header().lsn,
                    target_space_id,
                    replica_id: v.header().replica_id,
                    replicaset_id: value.replicaset_id.to_string(),
                    vshard_group: value.vshard_group.to_string(),
                };
                let (_, body) = v.destruct();
                let operation = Operation::Update {
                    ops: body.ops,
                    key: body.key,
                };
                DDL { meta, operation }
            }
            package::DML::Delete(v) => {
                let space_id =
                    u32::try_from(v.body().space_id).expect("casting u64 to u32 should not fail");
                if !super::is_system_space(space_id) {
                    log::info!(
                        "cannot cast DML to DDL cause its not system space with id: {space_id}"
                    );
                    return Err(package::DML::Delete(v));
                }
                if super::is_ignored_system_space(space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its DDL onto system space with id: {space_id}"
                    );
                    return Err(package::DML::Delete(v));
                }
                let target_space_id = u32::try_from(
                    v.body()
                        .key
                        .as_array()
                        .unwrap()
                        .first()
                        .unwrap()
                        .as_u64()
                        .unwrap(),
                )
                .expect("casting u64 to u32 should not fail");
                if super::is_system_space(target_space_id) {
                    log::info!(
                        "wont cast DML to DDL cause its DDL onto system space with id: {target_space_id}"
                    );
                    return Err(package::DML::Delete(v));
                }

                let meta = Meta {
                    space_id,
                    lsn: v.header().lsn,
                    target_space_id,
                    replica_id: v.header().replica_id,
                    replicaset_id: value.replicaset_id.to_string(),
                    vshard_group: value.vshard_group.to_string(),
                };
                let (_, body) = v.destruct();
                let operation = Operation::Delete { key: body.key };
                DDL { meta, operation }
            }
            operation => return Err(operation),
        };
        Ok(ddl)
    }
}

impl Into<coordinator::DDL> for &DDL {
    fn into(self) -> coordinator::DDL {
        let space_id = self.meta.space_id;
        let target_space_id = self.meta.target_space_id;
        let mut target_space_name = None;
        let ddl = match &self.operation {
            Operation::Insert { tuple } => {
                if space_id == config::INDEX_SPACE_ID {
                    let guard = unsafe {
                        conf::VSHARD_MAP
                            .get()
                            .expect("vshard map should be filled")
                            .read()
                            .expect("reading vshard map should not fail")
                    };
                    let cfg = guard
                        .get(&self.meta.vshard_group)
                        .expect("vshard group config should exist")
                        .get(&self.meta.replicaset_id)
                        .expect("replicaset config should exist")
                        .get(&target_space_id);
                    if cfg.is_none() {
                        target_space_name = None;
                    } else {
                        target_space_name = Some(cfg.unwrap().space_name.clone());
                    }
                }
                coordinator::SidecarDDL {
                    space_id,
                    key: vec![],
                    data: utils::value_to_vec(tuple),
                    kind: coordinator::SidecarDDLType::Insert,
                    target_space_name,
                }
            }
            Operation::Update { key, ops } => {
                if space_id == config::SPACE_SPACE_ID || space_id == config::INDEX_SPACE_ID {
                    let guard = unsafe {
                        conf::VSHARD_MAP
                            .get()
                            .expect("vshard map should be filled")
                            .read()
                            .expect("reading vshard map should not fail")
                    };
                    let cfg = guard
                        .get(&self.meta.vshard_group)
                        .expect("vshard group config should exist")
                        .get(&self.meta.replicaset_id)
                        .expect("replicaset config should exist")
                        .get(&target_space_id);
                    if cfg.is_none() {
                        target_space_name = None;
                    } else {
                        target_space_name = Some(cfg.unwrap().space_name.clone());
                    }
                }

                coordinator::SidecarDDL {
                    space_id,
                    key: utils::value_to_vec(key),
                    data: utils::value_to_vec(ops),
                    kind: coordinator::SidecarDDLType::Update,
                    target_space_name,
                }
            }
            Operation::Delete { key } => {
                if space_id == config::SPACE_SPACE_ID || space_id == config::INDEX_SPACE_ID {
                    let guard = unsafe {
                        conf::VSHARD_MAP
                            .get()
                            .expect("vshard map should be filled")
                            .read()
                            .expect("reading vshard map should not fail")
                    };
                    let cfg = guard
                        .get(&self.meta.vshard_group)
                        .expect("vshard group config should exist")
                        .get(&self.meta.replicaset_id)
                        .expect("replicaset config should exist")
                        .get(&target_space_id);
                    if cfg.is_none() {
                        target_space_name = None;
                    } else {
                        target_space_name = Some(cfg.unwrap().space_name.clone());
                    }
                }

                coordinator::SidecarDDL {
                    space_id,
                    key: utils::value_to_vec(key),
                    data: vec![],
                    kind: coordinator::SidecarDDLType::Delete,
                    target_space_name,
                }
            }
        };

        log::info!("converted DDL to sidecar for space_id: {space_id}, onto {:?} (space_id: {target_space_id})", ddl.target_space_name);
        coordinator::DDL {
            vshard_group: self.meta.vshard_group.to_string(),
            replicaset_id: self.meta.replicaset_id.to_string(),
            ddl,
        }
    }
}

impl Into<super::DML> for &std::rc::Rc<DDL> {
    fn into(self) -> super::DML {
        super::DML {
            meta: super::dml::Meta {
                bucket_id: None,
                lsn: self.meta.lsn,
                space_name: None,
                vshard_group: self.meta.vshard_group.clone(),
                replica_id: self.meta.replica_id,
                processing_kind: super::dml::ProcessingKind::Write,
            },
            operation: super::dml::Operation::Empty,
        }
    }
}
