use deepsize::DeepSizeOf;
use serde::{Deserialize, Serialize};

use picomole::package;

use crate::{config, error, sidecar};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum ProcessingKind {
    Store,
    Write,
    Commit,
}

#[derive(Debug)]
pub struct IRDML<'a> {
    dml: package::DML,
    replicaset_id: &'a str,
    vshard_group: &'a str,
}

impl<'a> IRDML<'a> {
    pub fn new(dml: package::DML, replicaset_id: &'a str, vshard_group: &'a str) -> Self {
        Self {
            dml,
            replicaset_id,
            vshard_group,
        }
    }
}

#[allow(clippy::unsafe_derive_deserialize)]
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Meta {
    pub bucket_id: Option<u64>,
    pub lsn: Option<u64>,
    pub space_name: Option<String>,
    pub vshard_group: String,
    pub replica_id: Option<u64>,
    pub processing_kind: ProcessingKind,
}

impl Meta {
    pub fn new(
        space_id: u32,
        lsn: Option<u64>,
        replica_id: Option<u64>,
        vshard_group: &str,
        replicaset_id: &str,
        header: &package::Header,
        tuple: &rmpv::Value,
    ) -> Self {
        let (bucket_id_field, space_name) = {
            let guard = unsafe {
                config::VSHARD_MAP
                    .get()
                    .expect("vshard map should be filled")
                    .read()
                    .expect("reading vshard map should not fail")
            };
            let (bucket_id_field, space_name) = match guard
                .get(vshard_group)
                .expect("vshard group config should exist")
                .get(replicaset_id)
                .expect("replicaset config should exist")
                .get(&space_id)
            {
                Some(cfg) => (Some(cfg.bucket_id_index), Some(cfg.space_name.clone())),
                None => (None, None),
            };
            drop(guard);
            (bucket_id_field, space_name)
        };

        let array = tuple.as_array().expect("tuple must be valid array");

        let is_tx = header.tsn.is_some();
        let is_last = header.flags.is_some();
        let is_sync = match header.flags {
            Some(v) => v & (1 << 0x02) != 0 || v & (1 << 0x04) != 0,
            None => false,
        };

        let processing_kind = if is_tx && !is_last || is_sync && is_last {
            ProcessingKind::Store
        } else if is_tx && is_last {
            ProcessingKind::Commit
        } else {
            ProcessingKind::Write
        };

        let bucket_id = bucket_id_field.map(|v| {
            array
                .get(v)
                .expect("bucket id must be in tuple")
                .as_u64()
                .unwrap_or_default()
        });

        Self {
            lsn,
            bucket_id,
            space_name,
            replica_id,
            processing_kind,
            vshard_group: vshard_group.to_string(),
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum Operation {
    Delete(rmpv::Value),
    Replace(rmpv::Value),
    Empty,
}

impl DeepSizeOf for Operation {
    fn deep_size_of_children(&self, context: &mut deepsize::Context) -> usize {
        match self {
            Operation::Delete(value) | Operation::Replace(value) => {
                size_of_rmpv_value(value, context)
            }
            Operation::Empty => 0,
        }
    }
}

#[allow(clippy::only_used_in_recursion)]
fn size_of_rmpv_value(v: &rmpv::Value, context: &mut deepsize::Context) -> usize {
    match v {
        rmpv::Value::Nil => 1,
        rmpv::Value::Boolean(v) => v.deep_size_of(),
        rmpv::Value::Integer(_) => 8,
        rmpv::Value::F64(v) => v.deep_size_of(),
        rmpv::Value::F32(v) => v.deep_size_of(),
        rmpv::Value::String(utf8_string) => utf8_string.as_bytes().deep_size_of(),
        rmpv::Value::Binary(vec) => vec.as_slice().deep_size_of(),
        rmpv::Value::Array(vec) => vec
            .iter()
            .fold(0, |acc, v| acc + size_of_rmpv_value(v, context)),
        rmpv::Value::Map(vec) => vec.iter().fold(0, |acc, (v1, v2)| {
            acc + size_of_rmpv_value(v1, context) + size_of_rmpv_value(v2, context)
        }),
        rmpv::Value::Ext(_, vec) => 1 + vec.as_slice().deep_size_of(),
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct DML {
    pub meta: Meta,
    pub operation: Operation,
}

impl DML {
    pub fn to_empty(&self) -> DML {
        DML {
            meta: self.meta.clone(),
            operation: Operation::Empty,
        }
    }

    pub fn is_empty(&self) -> bool {
        matches!(self.operation, Operation::Empty)
    }
}

impl<'a> TryFrom<IRDML<'a>> for DML {
    type Error = error::Error;
    fn try_from(value: IRDML) -> Result<Self, Self::Error> {
        log::info!("try to convert DML to DML: {value:?}");
        match value.dml {
            package::DML::Insert(v) => {
                let (header, body) = v.destruct();
                let tuple = body.tuple;
                let space_id =
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail");
                let meta = Meta::new(
                    space_id,
                    header.lsn,
                    header.replica_id,
                    value.vshard_group,
                    value.replicaset_id,
                    &header,
                    &tuple,
                );
                let operation = if super::is_system_space(space_id) || meta.space_name.is_none() {
                    Operation::Empty
                } else {
                    Operation::Replace(tuple)
                };
                Ok(DML { meta, operation })
            }
            package::DML::Update(v) => {
                let (header, body) = v.destruct();
                let tuple = body
                    .new_tuple
                    .expect("new tuple must be presented in update operation");
                let space_id =
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail");
                let meta = Meta::new(
                    space_id,
                    header.lsn,
                    header.replica_id,
                    value.vshard_group,
                    value.replicaset_id,
                    &header,
                    &tuple,
                );
                let operation = if super::is_system_space(space_id) || meta.space_name.is_none() {
                    Operation::Empty
                } else {
                    Operation::Replace(tuple)
                };
                Ok(DML { meta, operation })
            }
            package::DML::Delete(v) => {
                let (header, body) = v.destruct();
                let tuple = body
                    .old_tuple
                    .expect("old tuple must be presented in delete operation");
                let key = body.key;
                let space_id =
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail");
                let meta = Meta::new(
                    space_id,
                    header.lsn,
                    header.replica_id,
                    value.vshard_group,
                    value.replicaset_id,
                    &header,
                    &tuple,
                );
                let operation = if super::is_system_space(space_id) || meta.space_name.is_none() {
                    Operation::Empty
                } else {
                    Operation::Delete(key)
                };
                Ok(DML { meta, operation })
            }
            package::DML::Upsert(v) => {
                let (header, body) = v.destruct();
                let tuple = body
                    .new_tuple
                    .expect("new tuple must be presented in upsert operation");
                let space_id =
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail");
                let meta = Meta::new(
                    space_id,
                    header.lsn,
                    header.replica_id,
                    value.vshard_group,
                    value.replicaset_id,
                    &header,
                    &tuple,
                );
                let operation = if super::is_system_space(space_id) || meta.space_name.is_none() {
                    Operation::Empty
                } else {
                    Operation::Replace(tuple)
                };
                Ok(DML { meta, operation })
            }
            package::DML::Replace(v) => {
                let (header, body) = v.destruct();
                let tuple = body
                    .new_tuple
                    .expect("new tuple must be presented in replace operation");
                let space_id =
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail");
                let meta = Meta::new(
                    u32::try_from(body.space_id).expect("casting u64 to u32 should not fail"),
                    header.lsn,
                    header.replica_id,
                    value.vshard_group,
                    value.replicaset_id,
                    &header,
                    &tuple,
                );
                let operation = if super::is_system_space(space_id) || meta.space_name.is_none() {
                    Operation::Empty
                } else {
                    Operation::Replace(tuple)
                };
                Ok(DML { meta, operation })
            }
        }
    }
}

impl Into<sidecar::DML> for &DML {
    fn into(self) -> sidecar::DML {
        match &self.operation {
            Operation::Delete(v) => sidecar::DML::Delete(sidecar::DMLDelete {
                space: self.meta.space_name.as_ref().unwrap().clone(),
                key: utils::value_to_vec(v),
            }),
            Operation::Replace(v) => sidecar::DML::Replace(sidecar::DMLReplace {
                space: self.meta.space_name.as_ref().unwrap().clone(),
                tuple: utils::value_to_vec(v),
            }),
            Operation::Empty => unreachable!("empty operation should not be sent to sidecar"),
        }
    }
}
