use std::{rc, time};

use crate::{config as conf, error, types};

mod ddl;
pub mod dml;
mod synchro;

pub use ddl::DDL;
pub use dml::DML;

use ddl::IRDDL;
use dml::IRDML;

pub use synchro::{Operation as SynchroOperation, Synchro};

fn is_system_space(id: u32) -> bool {
    conf::SYSTEM_SPACES.contains(&id)
}
fn is_ignored_system_space(id: u32) -> bool {
    conf::IGNORED_SYSTEM_SPACES.contains(&id)
}

pub struct IR<'a> {
    package: picomole::ReplicationPackage,
    vshard_group: &'a str,
    replicaset_id: &'a str,
}

impl<'a> IR<'a> {
    pub fn new(
        package: picomole::ReplicationPackage,
        vshard_group: &'a str,
        replicaset_id: &'a str,
    ) -> Self {
        Self {
            package,
            vshard_group,
            replicaset_id,
        }
    }
}

#[derive(Debug, Clone)]
pub enum Operation {
    DML(rc::Rc<DML>),
    DDL(rc::Rc<DDL>),
    Syncro(rc::Rc<Synchro>),
    State(rc::Rc<types::State>),
}

#[derive(Debug, Clone)]
pub struct Package {
    pub operation: Operation,
    pub received_at: time::Instant,
}

unsafe impl Send for Package {}

impl Package {
    pub fn commit_info(&self) -> Option<(u64, u64)> {
        match &self.operation {
            Operation::DDL(v) => {
                if v.meta.lsn.is_some() && v.meta.replica_id.is_some() {
                    Some((v.meta.replica_id.unwrap(), v.meta.lsn.unwrap()))
                } else {
                    None
                }
            }
            Operation::DML(v) => {
                if v.meta.lsn.is_none() || v.meta.replica_id.is_none() {
                    None
                } else {
                    match v.meta.processing_kind {
                        dml::ProcessingKind::Store => None,
                        dml::ProcessingKind::Write | dml::ProcessingKind::Commit => {
                            Some((v.meta.replica_id.unwrap(), v.meta.lsn.unwrap()))
                        }
                    }
                }
            }
            Operation::Syncro(v) => Some((v.replica_id, v.lsn)),
            Operation::State(_) => None,
        }
    }
}

impl<'a> TryFrom<IR<'a>> for Package {
    type Error = error::Error;

    fn try_from(value: IR) -> Result<Self, Self::Error> {
        let dml = match value.package {
            picomole::ReplicationPackage::Synchro(syncro) => {
                return Ok(Self {
                    received_at: time::Instant::now(),
                    operation: Operation::Syncro(rc::Rc::new(Synchro::from(syncro))),
                })
            }
            picomole::ReplicationPackage::DML(dml) => dml,
        };
        let dml = match DDL::try_from(IRDDL::new(dml, value.replicaset_id, value.vshard_group)) {
            Ok(v) => {
                return Ok(Self {
                    received_at: time::Instant::now(),
                    operation: Operation::DDL(rc::Rc::new(v)),
                })
            }
            Err(e) => e,
        };
        let operation = DML::try_from(IRDML::new(dml, value.replicaset_id, value.vshard_group))?;
        Ok(Self {
            received_at: time::Instant::now(),
            operation: Operation::DML(rc::Rc::new(operation)),
        })
    }
}
