use std::{rc, time};

use picoplugin::transport::rpc;

use crate::{config, error, limbo, package, types};

pub struct Processor {
    limbo: limbo::Limbo,
    replica_info: rc::Rc<types::ReplicaInfo>,
    leader_id: String,
    routes: types::BucketsChannels,
}

impl Processor {
    pub fn new(
        leader_id: String,
        replica_info: rc::Rc<types::ReplicaInfo>,
        routes: types::BucketsChannels,
    ) -> Self {
        let limbo = limbo::Limbo::new();
        Self {
            limbo,
            replica_info,
            leader_id,
            routes,
        }
    }

    pub fn apply(&mut self, package: package::Package) -> Result<(), error::Error> {
        let received_at = package.received_at;
        match package.operation {
            package::Operation::DML(dml) => self.apply_dml(dml, received_at),
            package::Operation::DDL(ddl) => self.apply_ddl(ddl),
            package::Operation::Syncro(syncro) => self.apply_syncro(syncro),
            package::Operation::State(_) => unreachable!("state should be processed by worker"),
        }
    }

    fn apply_syncro(&mut self, syncro: rc::Rc<package::Synchro>) -> Result<(), error::Error> {
        match syncro.operation {
            package::SynchroOperation::Commit((_, lsn)) => {
                self.commit_transaction(lsn)?;
            }
            package::SynchroOperation::Rollback((_, lsn)) => {
                self.rollback_transaction(lsn);
            }
        };
        Ok(())
    }

    fn apply_ddl(&mut self, ddl: rc::Rc<package::DDL>) -> Result<(), error::Error> {
        log::info!("apply ddl: {ddl:?}");
        let operation: coordinator::DDL = (&*ddl).into();
        rpc::RequestBuilder::new(rpc::RequestTarget::InstanceId(&self.leader_id))
            .path(config::RPC_COORDINATOR_DDL_ROUTE)
            .plugin_service("uroboros", "uroboros")
            .plugin_version(env!("CARGO_PKG_VERSION"))
            .input(
                rpc::Request::encode_rmp(&operation).expect("encoding ddl to rmp should not fail"),
            )
            .send()
            .map_err(|e| {
                log::error!("failed to apply ddl: {e:?}",);
                error::Error::WriterApplyDDL
            })?;
        Ok(())
    }

    fn apply_dml(
        &mut self,
        dml: rc::Rc<package::DML>,
        received_at: time::Instant,
    ) -> Result<(), error::Error> {
        log::info!("apply dml: {dml:?}");
        match dml.meta.processing_kind {
            package::dml::ProcessingKind::Commit => {
                self.limbo.push(dml.clone(), received_at);
                self.commit_transaction(dml.meta.lsn.unwrap())?;
            }
            package::dml::ProcessingKind::Store => {
                self.limbo.push(dml.clone(), received_at);
            }
            package::dml::ProcessingKind::Write => self.write_dml(dml, received_at)?,
        }
        Ok(())
    }

    fn commit_transaction(&mut self, lsn: u64) -> Result<(), error::Error> {
        for (dml, received_at) in self.limbo.take_to(lsn) {
            self.write_dml(dml, received_at)?;
        }
        Ok(())
    }

    fn rollback_transaction(&mut self, lsn: u64) {
        drop(self.limbo.take_from(lsn));
    }

    fn write_dml(
        &self,
        dml: rc::Rc<package::DML>,
        received_at: time::Instant,
    ) -> Result<(), error::Error> {
        let meta = dml.as_ref().meta.clone();
        let bucket_id = usize::try_from(meta.bucket_id.unwrap())
            .expect("convert bucket_id to usize should not fail");
        self.routes
            .get(bucket_id - 1)
            .unwrap_or_else(|| {
                panic!(
                    "bucket {bucket_id} should exist in routes for vshard group {}",
                    meta.vshard_group,
                )
            })
            .send((
                self.replica_info.clone(),
                meta.replica_id.unwrap_or_default(),
                types::ReplicaMessage::DML(dml),
                received_at,
            ))
            .map_err(|e| {
                log::error!("failed to send DML to channel: {e:?}",);
                error::Error::BufferCreateRecord
            })?;
        Ok(())
    }
}
