use picoplugin::system::tarantool::{
    cbus::sync::tokio as cbus_tokio, fiber::r#async::futures::executor::block_on,
};
use std::{io, mem, num, rc, sync as s_sync, time as s_time};
use tokio::{
    runtime,
    sync::{mpsc, oneshot},
    task, time,
};
use tokio_util::sync as tu_sync;

use crate::{config, error, package, types};

type ReaderCommitSender = mpsc::Sender<(u64, u64)>;
type ReplicaCommitReceiver = mpsc::Receiver<(u64, u64)>;
type ReplicaCommitSender = cbus_tokio::Sender<(u64, u64)>;
type ReaderCommitReceiver = cbus_tokio::EndpointReceiver<(u64, u64)>;
type ReplicaStateSender = cbus_tokio::Sender<types::State>;
type ReaderStateReceiver = cbus_tokio::EndpointReceiver<types::State>;
type PackageSender = cbus_tokio::Sender<Result<package::Package, error::Error>>;
type PackageReceiver = cbus_tokio::EndpointReceiver<Result<package::Package, error::Error>>;

/// Represents a reader for exact replicaset.
pub struct Reader {
    /// Current state of reader which contains processed `LSN`.
    state: Option<types::State>,
    /// Handle of current [`runtime::Runtime`].
    rt_handle: runtime::Handle,
    /// Background running instance of [`Replica`].
    replica: mem::ManuallyDrop<task::JoinHandle<()>>,
    /// Receiver of replication packages from [`Replica`].
    package_receiver: mem::ManuallyDrop<PackageReceiver>,
    /// Receiver of commit packages from [`Replica`].
    commit_sender: mem::ManuallyDrop<ReaderCommitSender>,
    /// Receiver of commit confirmation from [`Replica`].
    commit_receiver: mem::ManuallyDrop<ReaderCommitReceiver>,
    /// Receiver of state after snapshot from [`Replica`].
    first_state_receiver: mem::ManuallyDrop<ReaderStateReceiver>,
    /// Receiver of state after bootstrap from [`Replica`].
    state_receiver: mem::ManuallyDrop<ReaderStateReceiver>,
    /// Token which is used to cancell reconnections as replica.
    connection_token: tu_sync::CancellationToken,
}

impl Reader {
    pub fn try_new(
        cfg: &config::ReaderConfig,
        rt_handle: runtime::Handle,
        instance_uuid: String,
    ) -> Result<Self, error::Error> {
        let connection_token = tu_sync::CancellationToken::new();
        let state = types::State::select(&cfg.replicaset_id)?;

        let (replica_package_sender, reader_package_receiver) = cbus_tokio::channel(
            config::READER_CBUS_ENDPOINT,
            num::NonZeroUsize::new(crate::READER_BUFFER_SIZE).unwrap(),
        );
        let (replica_commit_sender, reader_commit_receiver) = cbus_tokio::channel(
            config::READER_CBUS_ENDPOINT,
            num::NonZeroUsize::new(1).unwrap(),
        );
        let (replica_first_state_sender, reader_first_state_receiver) = cbus_tokio::channel(
            config::READER_CBUS_ENDPOINT,
            num::NonZeroUsize::new(1).unwrap(),
        );
        let (replica_state_sender, reader_state_receiver) = cbus_tokio::channel(
            config::READER_CBUS_ENDPOINT,
            num::NonZeroUsize::new(1).unwrap(),
        );
        let (reader_commit_sender, replica_commit_receiver) =
            mpsc::channel(crate::READER_BUFFER_SIZE);

        let replica = rt_handle.block_on(Replica::try_new(
            cfg.clone(),
            rt_handle.clone(),
            state.clone(),
            replica_package_sender,
            replica_commit_receiver,
            replica_commit_sender,
            replica_first_state_sender,
            replica_state_sender,
            instance_uuid,
            connection_token.clone(),
        ))?;
        let replica = rt_handle.spawn(replica.start());
        Ok(Self {
            state,
            rt_handle,
            connection_token,
            replica: mem::ManuallyDrop::new(replica),
            commit_sender: mem::ManuallyDrop::new(reader_commit_sender),
            commit_receiver: mem::ManuallyDrop::new(reader_commit_receiver),
            first_state_receiver: mem::ManuallyDrop::new(reader_first_state_receiver),
            state_receiver: mem::ManuallyDrop::new(reader_state_receiver),
            package_receiver: mem::ManuallyDrop::new(reader_package_receiver),
        })
    }

    pub fn read(&mut self) -> Result<package::Package, error::Error> {
        // Check if snapshot is ended and we need to save state
        if self.state.is_none() && !self.first_state_receiver.is_empty() {
            let state = self
                .first_state_receiver
                .receive()
                .expect("receiving from first state channel should not fail");
            self.state = Some(state.clone());
            return Ok(package::Package {
                operation: package::Operation::State(rc::Rc::new(state)),
                received_at: std::time::Instant::now(),
            });
        }
        // replication stream
        let package = self
            .package_receiver
            .receive()
            .expect("cbus receiver cannot be closed while call read on reader")?;
        // Check if we should update state from commit
        if let Some((replica_id, lsn)) = package.commit_info() {
            if self.state.is_none() {
                log::warn!(
                    "skip commit replica_id: {replica_id}, lsn: {lsn} cause bootstrap is not yet"
                );
                return Ok(package);
            }
            let state = self.state.as_mut().unwrap();
            block_on(async {
                self.commit_sender
                    .send((replica_id, lsn))
                    .await
                    .expect("sending commit to reader replica should not fail");
            });
            let (replica_id, lsn) = self
                .commit_receiver
                .receive()
                .expect("receiving commit from reader replica should not fail");
            state.vclock.insert(replica_id, lsn);
        }
        Ok(package)
    }

    pub fn is_empty(&self) -> bool {
        self.package_receiver.is_empty()
    }
}

impl Drop for Reader {
    fn drop(&mut self) {
        // SAFETY: this is safe cause it's the only place we call take
        unsafe {
            // First we drop commit channels so it's task will be able to join
            drop(mem::ManuallyDrop::take(&mut self.commit_sender));
            drop(mem::ManuallyDrop::take(&mut self.commit_receiver));
            // Secondly drop receiver channel so it's task will be able to join
            drop(mem::ManuallyDrop::take(&mut self.package_receiver));
            // Now we can drop actual replica reader
            self.rt_handle
                .block_on(mem::ManuallyDrop::take(&mut self.replica))
                .expect("joining replica should not fail");
            // Finally we drop state receivers so there will be no panic
            drop(mem::ManuallyDrop::take(&mut self.first_state_receiver));
            drop(mem::ManuallyDrop::take(&mut self.state_receiver));
        }
        // After we released all the resources we break all the reconnects
        self.connection_token.cancel();
    }
}

/// Represents an actual connected replica for exact replicaset.
struct Replica {
    /// Config to connect as replica.
    cfg: config::ReaderConfig,
    /// Vshard group which contains exact replicaset.
    vshard_group: String,
    /// Exact replicaset to receive data from.
    replicaset_id: String,
    /// Current instance uuid to use it as a replica id for replication
    instance_uuid: String,
    /// State with was used to start replica.
    state: Option<types::State>,
    /// Handle of current [`runtime::Runtime`].
    rt_handle: runtime::Handle,
    /// `IProto` connection to leader on replicaset.
    connection: Option<picomole::Connection>,
    /// Sender for received replication packages.
    package_sender: PackageSender,
    /// Sender of state after bootstrapping replication from snapshot.
    first_state_sender: Option<ReplicaStateSender>,
    /// Sender of state after bootstrapping replication from XLOG.
    state_sender: Option<ReplicaStateSender>,
    /// Sender of commited `replica_id - lsn` pairs.
    commit_sender: Option<ReplicaCommitSender>,
    /// Receiver of commited `replica_id - lsn` pairs.
    commit_receiver: Option<ReplicaCommitReceiver>,
    /// Token which is used to cancell reconnections as replica.
    connection_token: Option<tu_sync::CancellationToken>,
}

#[derive(Debug, Clone)]
struct ReplicaChannel {
    counter: usize,
    chan: tokio::sync::mpsc::Sender<picomole::ReplicationPackage>,
}

impl picomole::ReplicationChannel for ReplicaChannel {
    async fn send(&mut self, package: picomole::ReplicationPackage) {
        self.chan
            .send(package)
            .await
            .expect("sending package should not fail");
        self.counter += 1;
        if self.counter >= 200 {
            self.counter = 0;
            tokio::time::sleep(time::Duration::from_millis(100)).await;
        }
    }
}

impl Replica {
    #[allow(clippy::too_many_arguments)]
    async fn try_new(
        cfg: config::ReaderConfig,
        rt_handle: runtime::Handle,
        state: Option<types::State>,
        package_sender: PackageSender,
        commit_receiver: ReplicaCommitReceiver,
        commit_sender: ReplicaCommitSender,
        first_state_sender: ReplicaStateSender,
        state_sender: ReplicaStateSender,
        instance_uuid: String,
        connection_token: tu_sync::CancellationToken,
    ) -> Result<Self, error::Error> {
        let connection = picomole::Connection::new(&cfg.addr, &cfg.username, &cfg.password)
            .await
            .map_err(|e| {
                log::error!(
                    "failed to connect replicaset with config: {:?}, error: {:?}",
                    cfg,
                    e
                );
                error::Error::ReaderReplicaConnect
            })?;
        Ok(Self {
            cfg: cfg.clone(),
            state,
            rt_handle,
            package_sender,
            connection: Some(connection),
            vshard_group: cfg.vshard_group,
            first_state_sender: Some(first_state_sender),
            state_sender: Some(state_sender),
            replicaset_id: cfg.replicaset_id,
            instance_uuid,
            commit_sender: Some(commit_sender),
            commit_receiver: Some(commit_receiver),
            connection_token: Some(connection_token),
        })
    }

    async fn start(mut self) {
        log::warn!(
            "start Replica for vshard group {}, replicaset {}",
            self.cfg.vshard_group,
            self.cfg.replicaset_id,
        );
        let (close_sender, mut close_receiver) = oneshot::channel();
        let (handle_sender, mut handle_receiver) = mpsc::channel::<picomole::ReplicationHandle>(1);

        let (package_sender, mut package_receiver) = mpsc::channel(crate::READER_BUFFER_SIZE);

        let replication = {
            let token = self.connection_token.take().unwrap();
            let cfg = self.cfg.clone();
            let rt_handle = self.rt_handle.clone();
            let state = self.state.take();
            let mut connection = self.connection.take().unwrap();
            let commit_sender = self.commit_sender.take().unwrap();
            let first_state_sender = self.first_state_sender.take().unwrap();
            let state_sender = self.state_sender.take().unwrap();
            let mut commit_receiver = self.commit_receiver.take().unwrap();
            let snap_package_sender = ReplicaChannel {
                counter: 0,
                chan: package_sender.clone(),
            };
            let repl_package_sender = ReplicaChannel {
                counter: 0,
                chan: package_sender.clone(),
            };
            let replicaset_uuid = uuid::Uuid::parse_str(self.replicaset_id.as_str()).ok();
            let instance_uuid = uuid::Uuid::parse_str(self.instance_uuid.as_str())
                .expect("parse master uuid should not fail");

            self.rt_handle.spawn(async move {
                // In case we have no saved state for this exact replica
                // We need to start from snapshot
                let state = if let Some(v) = state {
                    state_sender
                        .send(v.clone())
                        .await
                        .expect("sending state should not fail");
                    v
                } else {
                    let (vclock, cluster_uuid) = loop {
                        match connection
                            .snapshot_channel(snap_package_sender.clone())
                            .await
                        {
                            Ok(v) => break v,
                            Err(e) => {
                                log::error!("failed to snapshot: {e:?}");
                                connection = if let Some(v) =
                                    connect_as_replica(&cfg, token.clone()).await
                                {
                                    v
                                } else {
                                    close_sender.send(()).expect(
                                        "sending to close inner replicator should not fail",
                                    );
                                    return Ok(());
                                };
                            }
                        }
                    };
                    assert_eq!(cluster_uuid, replicaset_uuid);
                    let state = types::State {
                        vclock,
                        replica_uuid: instance_uuid,
                        cluster_uuid: cluster_uuid.expect("cluster uuid must not be none"),
                    };
                    first_state_sender
                        .send(state.clone())
                        .await
                        .expect("sending state should not fail");
                    state
                };

                // Track vclock also locally so we can recreate connection
                let vclock = s_sync::Arc::new(s_sync::Mutex::new(state.vclock.clone()));

                let handler = {
                    let vclock = vclock.clone();
                    rt_handle.spawn(async move {
                        let mut handle = handle_receiver
                            .recv()
                            .await
                            .expect("receiving handle should not fail");

                        loop {
                            let message = tokio::select! {
                                h = handle_receiver.recv() => {
                                    handle = h.expect("receiving handle should not fail on select");
                                    continue
                                },
                                m = commit_receiver.recv() => m,
                            };
                            // Commit receiver channel was closed
                            let Some((replica_id, lsn)) = message else {
                                break;
                            };
                            handle.update_vclock(replica_id, lsn);
                            vclock.lock().unwrap().insert(replica_id, lsn);
                            commit_sender
                                .send((replica_id, lsn))
                                .await
                                .expect("sending commit should not fail");
                        }
                    })
                };
                tokio::pin!(handler);

                let res = loop {
                    let (replication, handle) = connection.subscribe_channel(
                        state.replica_uuid,
                        state.cluster_uuid,
                        vclock.lock().unwrap().clone(),
                        repl_package_sender.clone(),
                    );

                    handle_sender
                        .send(handle)
                        .await
                        .expect("sending handle should not fail");

                    let replicator = rt_handle.spawn(async move {
                        match replication.await {
                            Ok(()) => {
                                log::warn!("stop inner replicator");
                                Ok(())
                            }
                            Err(picomole::SubscribeError::Read(picomole::ReadError::IO(e)))
                                if e.kind() == io::ErrorKind::UnexpectedEof =>
                            {
                                // This also may happen in case if provided start lsn is not valid
                                // For existing XLOG files
                                // TODO: handle this uppercase
                                log::warn!("connection was closed from remote in inner replicator");
                                Err(error::Error::ReaderReplicaStopped)
                            }
                            Err(e) => {
                                log::error!("inner replicator is stopped unexpected: {e:?}");
                                Err(error::Error::ReaderReplicaClosed)
                            }
                        }
                    });

                    tokio::pin!(replicator);

                    tokio::select! {
                        // If replicator finished before handler - connection was broken and we need to recreate replication process
                        _ = &mut replicator => {
                            connection = match connect_as_replica(&cfg, token.clone()).await {
                                Some(v) => v,
                                None => break Ok(())
                            };
                        },
                        // Handler was finished first - so it's channel was closed and job just needs to stop regullary
                        handler_result = &mut handler => {
                            handler_result.expect("wait on handler should not fail");
                            break match replicator.await {
                                Ok(v) => v,
                                Err(e) => {
                                    log::error!("failed to join inner replicator: {e}");
                                    Ok(())
                                }
                            }
                        },
                    };
                };

                close_sender
                    .send(())
                    .expect("sending to close inner replicator should not fail");
                res
            })
        };

        loop {
            if close_receiver.try_recv().is_ok() {
                break;
            }
            let package = match package_receiver.try_recv() {
                Ok(v) => v,
                Err(e) => match e {
                    mpsc::error::TryRecvError::Empty => {
                        tokio::time::sleep(time::Duration::from_millis(100)).await;
                        continue;
                    }
                    mpsc::error::TryRecvError::Disconnected => {
                        break;
                    }
                },
            };
            if self
                .package_sender
                .send(package::Package::try_from(package::IR::new(
                    package,
                    &self.vshard_group,
                    &self.replicaset_id,
                )))
                .await
                .is_err()
            {
                if let Err(e) = replication
                    .await
                    .expect("joining replication should not fail")
                {
                    log::error!("failed on replication: {e:?}");
                };
                return;
            }
        }
        log::warn!(
            "stop Replica for vshard group {}, replicaset {}",
            self.cfg.vshard_group,
            self.cfg.replicaset_id,
        );
    }
}

async fn connect_as_replica(
    cfg: &config::ReaderConfig,
    token: tu_sync::CancellationToken,
) -> Option<picomole::Connection> {
    loop {
        tokio::select! {
            result = picomole::Connection::new(&cfg.addr, &cfg.username, &cfg.password) => {
                match result {
                    Ok(v) => {break Some(v)},
                    Err(e) => {
                        log::error!(
                            "failed to connect replicaset with config: {:?}, error: {:?}",
                            cfg,
                            e
                        );
                        time::sleep(s_time::Duration::from_millis(100)).await;
                        continue
                    }
                }
            },
            () = token.cancelled() => {break None}
        }
    }
}
