use std::{collections, rc};
use tokio::runtime;

use picoplugin::{
    internal,
    system::tarantool::{cbus, fiber},
};

use crate::{
    config, data::lsn_queue::LSNQueue, error, filter, lua, processor, reader, types, worker, writer,
};

pub struct Replicator<'a> {
    cbus_fiber_id: u64,
    is_running: bool,
    rt_handle: Option<runtime::Handle>,
    /// Collection of workers (one worker per replicaset) per vshard group
    workers: Option<collections::HashMap<String, collections::HashMap<String, worker::Handle<'a>>>>,
    /// Collection of all writers for current instance
    writers: Option<Vec<writer::TarantoolHandle<'a>>>,
    /// Collection of channels for send DML to writers (per vshard group)
    /// It's a vector of channels, vector index is bucket id
    routes: Option<collections::HashMap<String, types::BucketsChannels>>,
    /// Struct to commit LSN's after sending to consumer
    lsn_queue: LSNQueue<'a>,
    /// Collection of unique channels to writers, per vshard group
    writers_channels: Option<collections::HashMap<String, types::BucketsChannels>>,
    instance_uuid: String,
}

impl<'a> Replicator<'a> {
    /// Creates a new `Replicator` within one node.
    ///
    /// # Panics
    ///
    /// - In case of failing setting up cbus.
    /// - In case of failing starting cbus fiber.
    /// - In case of failing creating buffer.
    /// - In case of failing createing LSN queue.
    ///
    /// # Returns
    ///
    /// An instance of `Replicator`
    #[must_use]
    pub fn new() -> Self {
        log::warn!("initialize uroboros replicator");
        let cbus_fiber_id = {
            let oneshot = fiber::Channel::new(1);
            let channel = oneshot.clone();
            fiber::Builder::new()
                .name(config::READER_CBUS_ENDPOINT_F)
                .func(move || {
                    channel
                        .send(fiber::id())
                        .expect("sending cbus fiber id should not fail");
                    cbus::Endpoint::new(config::READER_CBUS_ENDPOINT)
                        .unwrap_or_else(|e| {
                            panic!(
                                "failed to setup cbus endpoint: {}, error: {e:?}.",
                                config::READER_CBUS_ENDPOINT,
                            )
                        })
                        .cbus_loop();
                })
                .start_non_joinable()
                .expect("starting cbus fiber should not fail");
            oneshot.recv().expect("recv cbus fiber id should not fail")
        };
        let info = internal::instance_info().expect("getting current info should not fail");
        lua::init();
        let lsn_queue = LSNQueue::new(fiber::Channel::new(crate::WRITER_BUFFER_SIZE));
        Self {
            cbus_fiber_id,
            workers: None,
            rt_handle: None,
            is_running: false,
            writers: None,
            routes: None,
            lsn_queue,
            writers_channels: None,
            instance_uuid: info.instance_uuid().into(),
        }
    }

    pub fn start(&mut self, rt_handle: runtime::Handle) {
        log::warn!("start uroboros replicator");
        if self.is_running {
            return;
        }
        self.rt_handle = Some(rt_handle);
        self.workers = Some(collections::HashMap::with_capacity(32));
        self.is_running = true;
    }

    pub fn configure(&mut self, cfg: &[config::VShardGroupConfig]) {
        unsafe { config::fill_vshard_map(cfg) };
    }

    /// Stop writer from current pool and recreate for new vshards.
    ///
    /// # Parameters
    ///
    /// - `writers_cfgs`: A configurations for a writers.
    /// - `buckets_per_writer`: Number of buckets per writer.
    ///
    /// # Panics
    ///
    /// - In case if there are no writers storage created.
    /// - In case if there are errors on init vshard for group.
    /// - In case if there are errors on writer create.
    ///
    pub fn create_writers(&mut self, writers_cfgs: &[config::WriterConfig]) {
        if self.writers.is_some() {
            drop(self.writers.take().expect("writers must exists on drop"));
        }
        let cap = writers_cfgs.len();
        self.writers_channels = Some(collections::HashMap::with_capacity(cap));
        self.routes = Some(collections::HashMap::with_capacity(cap));
        let mut writers_cap = 0;
        for cfg in writers_cfgs {
            match cfg {
                config::WriterConfig::Kafka(_) => {
                    unimplemented!("kafka writer is not implemented");
                }
                config::WriterConfig::Tarantool(tnt) => {
                    writers_cap +=
                        usize::try_from(tnt.vshard.bucket_count / tnt.buckets_per_writer + 1)
                            .expect("too small buckets_per_writer value");
                }
            }
        }
        let mut writers = Vec::with_capacity(writers_cap);

        for cfg in writers_cfgs {
            match cfg {
                config::WriterConfig::Kafka(_) => {
                    unimplemented!("kafka writer is not implemented");
                }
                config::WriterConfig::Tarantool(tnt) => {
                    lua::init_vshard(tnt.vshard.clone()).expect("init vshard should not fail");
                    self.routes
                        .as_mut()
                        .expect("writers must exist on add")
                        .insert(tnt.vshard.name.clone(), {
                            let mut fibers = Vec::with_capacity(
                                usize::try_from(tnt.vshard.bucket_count)
                                    .expect("convert capacity to usize should not fail"),
                            );
                            let mut writer_channels = Vec::with_capacity(writers_cap);
                            let mut first = 1;
                            while first < tnt.vshard.bucket_count {
                                let last = std::cmp::min(
                                    first + tnt.buckets_per_writer,
                                    tnt.vshard.bucket_count + 1,
                                );
                                let route_ch = fiber::Channel::new(crate::WRITER_BUFFER_SIZE);
                                let writer = writer::TarantoolWriter::new(
                                    tnt.vshard.name.clone(),
                                    first,
                                    last,
                                    route_ch.clone(),
                                    self.lsn_queue.channel(),
                                );
                                writers.push(writer.start());
                                for _i in first..last {
                                    fibers.push(route_ch.clone());
                                }
                                writer_channels.push(route_ch);
                                first += tnt.buckets_per_writer;
                            }
                            self.writers_channels
                                .as_mut()
                                .expect("writers channels must exists on add")
                                .insert(tnt.vshard.name.clone(), writer_channels);
                            fibers
                        });
                }
            }
        }
        self.writers = Some(writers);
    }

    /// Add a new replicator worker.
    ///
    /// # Parameters
    ///
    /// - `cfg`: A configuration for a replicator worker.
    ///
    /// # Panics
    ///
    /// - In case if there are no runtime handle.
    /// - In case if there are no workes storage created.
    ///
    /// # Errors
    ///
    /// - Return as error if failed to connect to a tarantool replica.
    ///
    /// # Returns
    ///
    /// A result with unit on success and error on fail.
    pub fn add(&mut self, cfg: config::JobConfig) -> Result<(), error::Error> {
        log::info!("add uroboros replicator worker: {cfg:?}");
        if !self.is_running {
            return Ok(());
        }
        let reader = reader::Reader::try_new(
            &cfg.reader,
            self.rt_handle
                .as_ref()
                .expect("runtime handle must exist on add")
                .clone(),
            self.instance_uuid.clone(),
        )?;
        let replica_info = rc::Rc::new(types::ReplicaInfo {
            vshard_group: cfg.vshard_group.clone(),
            replicaset_uuid: cfg.replicaset_id.clone(),
            replicaset_alias: cfg.replicaset_alias.clone(),
            replica_uuid: self.instance_uuid.clone(),
        });
        let processor = processor::Processor::new(
            cfg.leader_id.clone(),
            replica_info.clone(),
            self.routes
                .as_ref()
                .expect("routes should exist on worker start")
                .get(&cfg.vshard_group)
                .unwrap_or_else(|| {
                    panic!(
                        "routes for {} should exist on worker start",
                        &cfg.vshard_group,
                    )
                })
                .clone(),
        );
        let worker = worker::Worker::new(
            replica_info,
            reader,
            processor,
            vec![
                Box::new(filter::Shard),
                Box::new(filter::Space::new(cfg.disabled_spaces)),
            ],
            self.writers_channels
                .as_ref()
                .expect("writers channels should exist on worker start")
                .get(&cfg.vshard_group)
                .unwrap_or_else(|| {
                    panic!(
                        "writers channels for {} should exist on worker start",
                        &cfg.vshard_group,
                    )
                })
                .clone(),
            self.lsn_queue.channel(),
        );
        if self
            .workers
            .as_ref()
            .expect("workers must exist on add")
            .get(&cfg.vshard_group)
            .is_none()
        {
            self.workers
                .as_mut()
                .expect("workers must exist on add")
                .insert(
                    cfg.vshard_group.clone(),
                    collections::HashMap::with_capacity(32),
                );
        }
        self.workers
            .as_mut()
            .expect("workers must exist on add")
            .get_mut(&cfg.vshard_group)
            .expect("workers must exist on add")
            .insert(cfg.replicaset_id.clone(), worker.start());
        Ok(())
    }

    /// Removes replicator for exact replicaset within exact vshard group, stopping it's process.
    ///
    /// # Parameters
    ///
    /// - `vshard_group`: A name of vshard group.
    /// - `replicaset_id`: An ID of replicaset.
    ///
    /// # Panics
    ///
    /// - In case if there are no vshard group.
    /// - In case if there are no replicaset within vshard group.
    pub fn remove(&mut self, vshard_group: &str, replicaset_id: &str) {
        log::info!("remove uroboros replicator worker with vshard group: {vshard_group}, replicaset_id: {replicaset_id}");
        if !self.is_running {
            return;
        }
        self.workers
            .as_mut()
            .expect("workers should exist on remove")
            .get_mut(vshard_group)
            .expect("vshard group workers should exist no remove")
            .remove(replicaset_id);
    }

    /// Stops entire replicator process, stopping each worker.
    ///
    /// # Panics
    ///
    /// - May panic in case of there are not workers, e.g. replicator was not started
    ///   which seems to be unreachable due to design.
    pub fn stop(&mut self) {
        log::warn!("stop uroboros replicator");
        if !self.is_running {
            return;
        }
        if self.routes.is_some() {
            drop(self.routes.take().expect("routes should exist on stop"));
        }
        if self.writers_channels.is_some() {
            drop(
                self.writers_channels
                    .take()
                    .expect("writers channels should exist on stop"),
            );
        }
        if self.writers.is_some() {
            drop(self.writers.take().expect("writers must exists on stop"));
        }
        if self.workers.is_some() {
            drop(self.workers.take().expect("workers must exists on stop"));
        }
        self.rt_handle = None;
        self.is_running = false;
    }
}

impl<'a> Drop for Replicator<'a> {
    fn drop(&mut self) {
        fiber::cancel(self.cbus_fiber_id);
        loop {
            if !fiber::exists(self.cbus_fiber_id) {
                break;
            }
            fiber::r#yield().expect("replicator should not be cancelled");
        }
    }
}
