use std::{mem, rc, time};

use picoplugin::system::tarantool::fiber;

use crate::{data::lsn_queue, filter, metrics, package, processor, reader, types};

pub struct Worker {
    replica_info: rc::Rc<types::ReplicaInfo>,
    reader: reader::Reader,
    processor: processor::Processor,
    filters: Vec<Box<dyn filter::Filter>>,
    writers_channels: types::BucketsChannels,
    lsn_ch: fiber::Channel<lsn_queue::LSNQueueMsg>,
}

pub struct Handle<'a> {
    handle: mem::ManuallyDrop<fiber::JoinHandle<'a, ()>>,
    close: fiber::Channel<()>,
}

impl Worker {
    pub fn new(
        replica_info: rc::Rc<types::ReplicaInfo>,
        reader: reader::Reader,
        processor: processor::Processor,
        filters: Vec<Box<dyn filter::Filter>>,
        writers_channels: types::BucketsChannels,
        lsn_ch: fiber::Channel<lsn_queue::LSNQueueMsg>,
    ) -> Self {
        Self {
            replica_info,
            reader,
            processor,
            filters,
            writers_channels,
            lsn_ch,
        }
    }

    pub fn start<'a>(mut self) -> Handle<'a> {
        log::warn!(
            "start worker for vshard group: {}, replicaset: {}",
            self.replica_info.vshard_group,
            self.replica_info.replicaset_uuid
        );
        let vhsard_group = self.replica_info.vshard_group.clone();
        let replicaset_id = self.replica_info.replicaset_uuid.clone();
        let close = fiber::Channel::new(1);
        let handle = {
            let close = close.clone();
            fiber::Builder::new()
                .func(move || {
                    let mut count = 0;
                    let mut prev_empty: Option<rc::Rc<package::DML>> = None;
                    let mut last_empty: Option<time::Instant> = None;
                    loop {
                        if close.try_recv().is_ok() {
                            log::info!(
                                "worker received stop signal with vshard group: {}, replicaset: {}",
                                self.replica_info.vshard_group,
                                self.replica_info.replicaset_uuid
                            );
                            return;
                        }
                        if self.reader.is_empty() {
                            // no new packages - let's try to store LSN from last empty package if it exists
                            if prev_empty.is_some() {
                                let since_empty = if let Some(v) = last_empty {
                                    v.elapsed()
                                } else {
                                    time::Duration::from_secs(0)
                                };
                                if since_empty.as_secs() >= 3 {
                                    let dml = prev_empty.unwrap();
                                    self.set_lsn(
                                        dml.meta.replica_id.unwrap(),
                                        dml.meta.lsn.unwrap(),
                                    );
                                    prev_empty = None;
                                    last_empty = None;
                                }
                            }
                            count = 0;
                            fiber::sleep(time::Duration::from_millis(100));
                            continue;
                        }
                        let read = self.reader.read();
                        log::info!("received from reader: {read:?}");
                        let mut package = match read {
                            Ok(v) => v,
                            Err(_e) => {
                                // We assume that any error on read is error on connection with master of replicaset
                                // So we just finish this replication process
                                // This state of master will be detected on coordinator while pinging master
                                // And this will drop current worker and create a new one for maybe a new master of replicaset
                                return;
                            }
                        };
                        let received_at = package.received_at;
                        metrics::WORKER_RECEIVE_LAG_GAUGE
                            .with_label_values(&[
                                &self.replica_info.vshard_group,
                                &self.replica_info.replicaset_uuid,
                                &self.replica_info.replicaset_alias.clone().unwrap_or_default(),
                            ])
                            .set(received_at.elapsed().as_secs_f64());
                        metrics::RECEIVED_PACKAGES_COUNTER
                            .with_label_values(&[
                                &self.replica_info.vshard_group,
                                &self.replica_info.replicaset_uuid,
                                &self.replica_info.replicaset_alias.clone().unwrap_or_default(),
                            ])
                            .inc();

                        for filter in &self.filters {
                            package = filter.apply(package);
                        }

                        metrics::RECEIVED_PACKAGES_FILTERED_COUNTER
                            .with_label_values(&[
                                &self.replica_info.vshard_group,
                                &self.replica_info.replicaset_uuid,
                                &self.replica_info.replicaset_alias.clone().unwrap_or_default(),
                            ])
                            .inc();

                        let received_at = package.received_at;

                        match package.operation {
                            package::Operation::State(state) => {
                                log::info!("package with state: {state:?}");
                                // we need to store vclock after full snapshot receiving
                                // so let wait until all DML's from this snapshot are processed
                                // and then store vclock
                                wait_state(self.replica_info.clone(), &self.writers_channels);
                                utils::tx_retry!(crate::data::state::create(&state), delay: 100, tries: 5, backoff: 2.0)
                                    .expect("state saving should not fail");
                            }
                            package::Operation::DML(dml) if dml.is_empty() => {
                                if let Some(_lsn) = dml.meta.lsn {
                                    prev_empty = Some(dml.clone());
                                    last_empty = Some(time::Instant::now());
                                }
                            }
                            _ => {
                                log::info!("package to apply: {package:?}");
                                // if there is last_empty (non stored LSN) we can skip it
                                // cause LSN will be stored after DML
                                prev_empty = None;
                                last_empty = None;
                                // DDL also may have LSN so we should save it
                                if let package::Operation::DDL(ddl) = &package.operation {
                                    if ddl.meta.lsn.is_some() {
                                        prev_empty = Some(rc::Rc::new(ddl.into()));
                                        last_empty = Some(time::Instant::now());
                                    }
                                }
                                self.processor
                                    .apply(package)
                                    .expect("applying package should not fail");
                            }
                        }
                        metrics::WORKER_APPLY_LAG_GAUGE
                            .with_label_values(&[
                                &self.replica_info.vshard_group,
                                &self.replica_info.replicaset_uuid,
                                &self.replica_info.replicaset_alias.clone().unwrap_or_default(),
                            ])
                            .set(received_at.elapsed().as_secs_f64());
                        log::info!("committed lsn");
                        count += 1;
                        if count >= 200 {
                            count = 0;
                            fiber::sleep(time::Duration::from_millis(100));
                        }
                    }
                })
                .name(format!("uroboros_worker_{vhsard_group}_{replicaset_id}"))
                .start()
                .expect("spawning worker fiber should not fail")
        };
        Handle {
            close,
            handle: mem::ManuallyDrop::new(handle),
        }
    }

    fn set_lsn(&self, replica_id: u64, lsn: u64) {
        // we need to store LSN for skipped operations
        // so let wait until all DML's from this snapshot are processed
        // and then store vclock
        wait_state(self.replica_info.clone(), &self.writers_channels);
        self.lsn_ch
            .send(lsn_queue::LSNQueueMsg::Written {
                replica_info: self.replica_info.clone(),
                replica_id,
                lsn,
            })
            .expect("send lsn data to channel should not fail");
    }
}

/// Create new channel, send it to all writers and wait until all writers send message to this channel
fn wait_state(replica_info: rc::Rc<types::ReplicaInfo>, writers: &types::BucketsChannels) {
    let mut count = writers.len();
    let state_ch: fiber::Channel<()> = fiber::Channel::new(1);
    for ch in writers {
        ch.send((
            replica_info.clone(),
            0,
            types::ReplicaMessage::State(state_ch.clone()),
            time::Instant::now(),
        ))
        .expect("send state message should not fail");
    }
    while count > 0 {
        if state_ch.is_empty() {
            fiber::sleep(std::time::Duration::from_millis(100));
        }
        if state_ch.recv().is_some() {
            count -= 1;
        }
    }
}

impl<'a> Drop for Handle<'a> {
    fn drop(&mut self) {
        self.close
            .send(())
            .expect("sending to close channel should not fail");
        unsafe {
            mem::ManuallyDrop::take(&mut self.handle).join();
        }
    }
}
