use std::mem;
use std::ops::Range;
use std::rc;
use std::time;

use deepsize::DeepSizeOf;
use picoplugin::system::tarantool::fiber;

use crate::data::lsn_queue::LSNQueueMsg;
use crate::{error, lua, metrics, package::DML, types};

pub struct Writer {
    vshard_group: String,
    buckets: Range<u64>,
    dml_ch: types::WriterChannel,
    lsn_ch: fiber::Channel<LSNQueueMsg>,
}

impl Writer {
    pub fn new(
        vshard_group: String,
        start_bucket: u64,
        end_bucket: u64,
        dml_ch: types::WriterChannel,
        lsn_ch: fiber::Channel<LSNQueueMsg>,
    ) -> Self {
        Self {
            vshard_group,
            buckets: Range {
                start: start_bucket,
                end: end_bucket,
            },
            dml_ch,
            lsn_ch,
        }
    }

    #[allow(clippy::cast_possible_truncation)]
    #[allow(clippy::cast_sign_loss)]
    pub fn start<'a>(self) -> Handle<'a> {
        let start_bucket = self.buckets.start;
        let end_bucket = self.buckets.end;
        log::info!(
            "start writer for vshard group: {}, buckets range: {start_bucket} - {end_bucket}",
            self.vshard_group.clone(),
        );
        let name = format!(
            "uroboros_writer_{}_{start_bucket}_{end_bucket}",
            self.vshard_group.clone(),
        );
        let close = fiber::Channel::new(1);
        let handle = {
            let close = close.clone();
            fiber::Builder::new()
                .func(move || {
                    let mut count = 0;
                    loop {
                        if close.try_recv().is_ok() {
                            log::info!(
                                "writer received stop signal with vshard group: {}, buckets range: {start_bucket} - {end_bucket}",
                                &self.vshard_group,
                            );
                            return;
                        }
                        if self.dml_ch.is_empty() {
                            count = 0;
                            fiber::sleep(time::Duration::from_millis(100));
                            continue;
                        }
                        if let Some((replica_info, replica_id, rm, received_at)) = self.dml_ch.recv() {
                            match rm {
                                types::ReplicaMessage::DML(op) => {
                                    let mut op_lsn = 0;
                                    if let Some(lsn) = op.meta.lsn {
                                        op_lsn = lsn;
                                        self.lsn_ch.send(LSNQueueMsg::Store{
                                            replica_info: replica_info.clone(),
                                            replica_id,
                                            lsn,
                                        })
                                        .expect("send lsn data to channel should not fail");
                                    }
                                    let vshard_group = replica_info.vshard_group.clone();
                                    let replicaset_uuid = replica_info.replicaset_uuid.clone();
                                    let replicaset_alias = replica_info
                                        .replicaset_alias
                                        .clone()
                                        .unwrap_or_default();
                                    let op_size = op.operation.deep_size_of() as u64;
                                    loop {
                                        let start = time::Instant::now();
                                        match utils::tx_retry!(write_dml(op.clone()), delay: 100, tries: 5, backoff: 2.0) {
                                            Ok(()) if op_lsn == 0 => {
                                                metrics::VSHARD_APPLY_GAUGE
                                                .with_label_values(&[
                                                    &vshard_group,
                                                    &replicaset_uuid,
                                                    &replicaset_alias,
                                                ])
                                                .set(start.elapsed().as_secs_f64());
                                                break;
                                            }
                                            Ok(()) => {
                                                self.lsn_ch.send(LSNQueueMsg::Written {
                                                    replica_info,
                                                    replica_id,
                                                    lsn: op_lsn,
                                                })
                                                .expect("send lsn data to channel should not fail");
                                                metrics::VSHARD_APPLY_GAUGE
                                                    .with_label_values(&[
                                                        &vshard_group,
                                                        &replicaset_uuid,
                                                        &replicaset_alias,
                                                    ])
                                                    .set(start.elapsed().as_secs_f64());
                                                break;
                                            }
                                            Err(e) => {
                                                log::error!(
                                                    "writer can not delete applied DML (vshard group {}): {e:?}",
                                                    replica_info.vshard_group.clone(),
                                                );
                                                fiber::sleep(time::Duration::from_millis(100));
                                            }
                                        }
                                    }
                                    metrics::APPLIED_PACKAGES_COUNTER
                                        .with_label_values(&[
                                            &vshard_group,
                                            &replicaset_uuid,
                                            &replicaset_alias,
                                        ])
                                        .inc();
                                    metrics::APPLIED_BYTES_COUNTER
                                        .with_label_values(&[
                                            &vshard_group,
                                            &replicaset_uuid,
                                            &replicaset_alias,
                                        ])
                                        .inc_by(op_size);
                                    metrics::LAG_GAUGE
                                        .with_label_values(&[
                                            &vshard_group,
                                            &replicaset_uuid,
                                            &replicaset_alias,
                                        ])
                                        .set(received_at.elapsed().as_secs_f64());
                                    // log::info!("committed lsn");
                                }
                                types::ReplicaMessage::State(chan) => chan.send(()).expect("send to state channel should not fail"),
                            }
                        }
                        count += 1;
                        if count >= 200 {
                            count = 0;
                            fiber::sleep(time::Duration::from_millis(100));
                        }
                    }
                })
                .name(name)
                .start()
                .expect("spawning worker fiber should not fail")
        };
        Handle {
            close,
            handle: mem::ManuallyDrop::new(handle),
        }
    }
}

fn write_dml(dml: rc::Rc<DML>) -> Result<(), error::Error> {
    let group = dml.meta.vshard_group.clone();
    lua::apply_dml(dml).map_err(|e| {
        log::error!("writer can not apply DML (vshard group {group}): {e:?}");
        e
    })
}

pub struct Handle<'a> {
    handle: mem::ManuallyDrop<fiber::JoinHandle<'a, ()>>,
    close: fiber::Channel<()>,
}

impl<'a> Drop for Handle<'a> {
    fn drop(&mut self) {
        self.close
            .send(())
            .expect("sending to close channel should not fail");
        unsafe {
            mem::ManuallyDrop::take(&mut self.handle).join();
        }
    }
}
