/// Converts `rmpv::Value` to byte vector.
///
/// # Panics
///
/// - In case of failed encoding, but used internally so this is invariant.
#[must_use]
pub fn value_to_vec(value: &rmpv::Value) -> Vec<u8> {
    let mut result = Vec::new();
    rmpv::encode::write_value(&mut result, value).expect("encode value should not fail");
    result
}

/// Converts byte vector to `rmpv::Value`.
///
/// # Panics
///
/// - In case of failed decoding, but used internally so this is invariant.

#[must_use]
pub fn vec_to_value(v: Vec<u8>) -> rmpv::Value {
    let mut cursor = std::io::Cursor::new(v);
    rmpv::decode::read_value(&mut cursor).expect("decode value should not fail")
}

#[macro_export]
macro_rules! tx_retry {
    ($operation:expr) => {
        loop {
            match $operation {
                Ok(result) => break result,
                Err(err) => {
                    log::error!("failed on infinite operation retry: {:?}", err);
                    continue;
                },
            }
        }
    };
    ($operation:expr, delay: $sleep_ms:expr) => {
        loop {
            match $operation {
                Ok(result) => break result,
                Err(err) => {
                    log::error!(
                        "failed on infinite operation retry with delay: {}ms, error: {:?}",
                        $sleep_ms,
                        err
                    );
                    picoplugin::system::tarantool::fiber::sleep(std::time::Duration::from_millis($sleep_ms));
                },
            }
        }
    };
    ($operation:expr, delay: $sleep_ms:expr, tries: $max_retries:expr) => {{
        let mut retries = 0;
        loop {
            match $operation {
                Ok(result) => break Ok(result),
                Err(err) => {
                    ::log::error!(
                        "failed on limited operation retry with tries: {}, delay: {}ms,  error: {:?}",
                        $max_retries,
                        $sleep_ms,
                        err
                    );
                    retries += 1;
                    if retries >= $max_retries {
                        break Err(err);
                    }
                    picoplugin::system::tarantool::fiber::sleep(std::time::Duration::from_millis($sleep_ms));
                },
            }
        }
    }};
    ($operation:expr, delay: $sleep_ms:expr, tries: $max_retries:expr, backoff: $backoff:expr) => {{
        let mut retries = 0;
        let mut current_delay = $sleep_ms;
        loop {
            match $operation {
                Ok(result) => break Ok(result),
                Err(err) => {
                    ::log::error!(
                        "failed on limited operation retry with tries: {}, delay: {}ms, backoff: {}x, error: {:?}",
                        $max_retries,
                        current_delay,
                        $backoff,
                        err
                    );
                    retries += 1;
                    if retries >= $max_retries {
                        break Err(err);
                    }
                    picoplugin::system::tarantool::fiber::sleep(std::time::Duration::from_millis(current_delay));
                    current_delay = (current_delay as f64 * $backoff).min(std::u64::MAX as f64) as u64;
                },
            }
        }
    }};
}
