#![allow(dead_code)]
#![allow(clippy::new_without_default)]
#![allow(clippy::items_after_statements)]
#![allow(clippy::blocks_in_conditions)]

use once_cell::{sync, unsync};
use std::time;
use tokio::runtime;

use picoplugin::{
    internal,
    plugin::{
        self,
        interface::{self, ErrorBox},
        prelude::{self, service_registrar},
    },
    system::tarantool::{fiber, log as t_log},
    transport::rpc,
};

const TOKIO_THREAD_NAME: &str = "uroboros-tokio";
const FORBIDDEN_CHARS: &str = "@:";

thread_local! {
    static REPLICATOR: unsync::Lazy<fiber::Mutex<replicator::Replicator<'static>>> = unsync::Lazy::new(|| fiber::Mutex::new(replicator::Replicator::new()));
    static COORDINATOR: unsync::Lazy<fiber::Mutex<coordinator::Coordinator<'static>>> = unsync::Lazy::new(|| fiber::Mutex::new(coordinator::Coordinator::new()));
}

static LOGGER: sync::Lazy<t_log::TarantoolLogger> = sync::Lazy::new(|| {
    t_log::TarantoolLogger::with_mapping(|level: log::Level| match level {
        log::Level::Error => t_log::SayLevel::Error,
        log::Level::Warn => t_log::SayLevel::Warn,
        log::Level::Info => t_log::SayLevel::Info,
        log::Level::Debug => t_log::SayLevel::Verbose,
        log::Level::Trace => t_log::SayLevel::Debug,
    })
});

/// Setups a new logger.
///
/// # Panics
///
/// - In case of failed logger set.
fn init_logger() {
    log::set_logger(&*LOGGER).map_or_else(
        |e| println!("failed to setup logger: {e:?}"),
        |()| log::set_max_level(log::LevelFilter::Trace),
    );
}

pub struct Uroboros {
    is_master: bool,
    config: config::Config,
    runtime: runtime::Runtime,
}

impl Uroboros {
    /// Creates a new unconfigrued instance of `Uroboros`.
    ///
    /// # Panics
    ///
    /// - In case of failed tokio runtime spawn.
    ///
    /// # Returns
    ///
    /// An instance of `Uroboros`.
    #[must_use]
    pub fn new() -> Self {
        init_logger();
        log::warn!("initialize uroboros instance");
        let is_master = false;
        let runtime = runtime::Builder::new_multi_thread()
            .worker_threads(3)
            .thread_name(TOKIO_THREAD_NAME)
            .enable_all()
            .build()
            .expect("spawn tokio runtime should not fail");
        let config = config::Config::default();
        Self {
            is_master,
            config,
            runtime,
        }
    }

    /// Applies a new config
    ///
    /// # Parameters
    ///
    /// - `config`: A configuration for a plugin.
    /// - `is_master`: True if the current instance is master
    ///
    /// # Errors
    ///
    /// - Return as error if password contains invalud characters.
    ///
    /// # Returns
    ///
    /// A result with unit on success and error on fail.
    pub fn configure(&mut self, config: config::Config, is_master: bool) -> Result<(), ErrorBox> {
        log::warn!("configure uroboros instance");
        if !config
            .password
            .chars()
            .all(|x| FORBIDDEN_CHARS.find(x).is_none())
        {
            return Err(format!(
                "user password contains invalid characters (some of '{FORBIDDEN_CHARS}')",
            )
            .into());
        }
        unsafe { config::PASSWORD.clone_from(&config.password) }
        self.config = config;
        self.is_master = is_master;
        Ok(())
    }

    pub fn rt_handle(&self) -> runtime::Handle {
        self.runtime.handle().clone()
    }

    /// Starts an `Uroboros` plugin.
    ///
    /// # Parameters
    ///
    /// - `ctx`: A picodata context
    ///
    /// # Panics
    ///
    /// - In case of registering RPC route fails.
    pub fn start(&mut self, ctx: &prelude::PicoContext) {
        log::warn!("start uroboros instance");
        ctx.metrics_collection().append(replicator::collect_metrics);
        log::warn!("setup uroboros rpc routes");
        rpc::RouteBuilder::from_pico_context(ctx)
            .path(config::RPC_COORDINATOR_DDL_ROUTE)
            .register(|req, _ctx| {
                let ddl = req.decode_rmp::<coordinator::DDL>()?;
                let cond = COORDINATOR.with(|x| {
                    let mut rf = x.lock();
                    rf.handle_ddl(ddl)
                });
                cond.wait();
                rpc::Response::encode_rmp(&())
            })
            .expect("registering coordinator ddl should not fail");
        rpc::RouteBuilder::from_pico_context(ctx)
            .path(config::RPC_REPLICATOR_ADD_ROUTE)
            .register(|req, _ctx| {
                let cfg = req.decode_rmp::<config::JobConfig>()?;
                let result = REPLICATOR.with(|x| {
                    let mut rf = x.lock();
                    rf.add(cfg)
                });
                rpc::Response::encode_rmp(&result)
            })
            .expect("registering replicator add should not fail");
        rpc::RouteBuilder::from_pico_context(ctx)
            .path(config::RPC_REPLICATOR_CONFIGURE_ROUTE)
            .register(|req, _ctx| {
                let cfg = req.decode_rmp::<Vec<config::VShardGroupConfig>>()?;
                REPLICATOR.with(|x| {
                    let mut rf = x.lock();
                    rf.configure(&cfg);
                });
                rpc::Response::encode_rmp(&())
            })
            .expect("registering replicator configure should not fail");
        rpc::RouteBuilder::from_pico_context(ctx)
            .path(config::RPC_REPLICATOR_REMOVE_ROUTE)
            .register(|req, _ctx| {
                let (vhsard_group, replicaset_id) = req.decode_rmp::<(String, String)>()?;
                REPLICATOR.with(|x| {
                    let mut rf = x.lock();
                    rf.remove(&vhsard_group, &replicaset_id);
                });
                rpc::Response::encode_rmp(&())
            })
            .expect("registering replicator remove should not fail");
        rpc::RouteBuilder::from_pico_context(ctx)
            .path(config::RPC_REPLICATOR_WRITERS_ROUTE)
            .register(|req, _ctx| {
                let writers_cfgs = req.decode_rmp::<Vec<config::WriterConfig>>()?;
                REPLICATOR.with(|x| {
                    let mut rf = x.lock();
                    rf.create_writers(&writers_cfgs);
                });
                rpc::Response::encode_rmp(&())
            })
            .expect("registering replicator configure should not fail");
        let handle = self.rt_handle();
        REPLICATOR.with(|x| {
            let mut n = x.lock();
            n.start(handle);
        });
        if self.is_master {
            log::warn!("current uroboros instance is leader, will setup coordinator");
            let cfg = self.config.clone();
            fiber::Builder::new()
                .func(move || {
                    log::warn!("start coordinator setup fiber, wait for 3 secs until start");
                    fiber::sleep(time::Duration::from_secs(10));
                    let info =
                        internal::instance_info().expect("getting leader info should not fail");
                    COORDINATOR.with(|x| {
                        let mut rf = x.lock();
                        rf.configure(cfg, info.instance_id().to_string());
                        rf.start();
                    });
                })
                .start_non_joinable()
                .expect("starting coordinator setup fiber should not fail");
        }
    }

    pub fn stop(&mut self) {
        log::warn!("stop uroboros instance");
        REPLICATOR.with(|x| {
            let mut n = x.lock();
            n.stop();
        });
        if self.is_master {
            COORDINATOR.with(|x| {
                let mut rf = x.lock();
                rf.stop();
            });
        }
    }
}

impl prelude::Service for Uroboros {
    type Config = config::Config;

    fn on_config_change(
        &mut self,
        ctx: &prelude::PicoContext,
        cfg: Self::Config,
        _: Self::Config,
    ) -> plugin::interface::CallbackResult<()> {
        log::warn!("uroboros instance received on_config_change signal");
        self.stop();
        self.configure(cfg, ctx.is_master())?;
        self.start(ctx);
        Ok(())
    }

    fn on_start(
        &mut self,
        ctx: &prelude::PicoContext,
        cfg: Self::Config,
    ) -> plugin::interface::CallbackResult<()> {
        log::warn!("uroboros instance received on_start signal");
        self.configure(cfg, ctx.is_master())?;
        self.start(ctx);
        Ok(())
    }

    fn on_leader_change(
        &mut self,
        ctx: &prelude::PicoContext,
    ) -> plugin::interface::CallbackResult<()> {
        log::warn!("uroboros instance received on_leader_change signal");
        self.stop();
        self.configure(self.config.clone(), ctx.is_master())?;
        self.start(ctx);
        Ok(())
    }

    fn on_stop(&mut self, _: &prelude::PicoContext) -> plugin::interface::CallbackResult<()> {
        log::warn!("uroboros instance received on_stop signal");
        self.stop();
        Ok(())
    }
}

#[service_registrar]
pub fn service_registrar(reg: &mut interface::ServiceRegistry) {
    reg.add("uroboros", env!("CARGO_PKG_VERSION"), Uroboros::new);
}
